from unittest import TestCase
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.ray_2d import Ray2D


class TestRay2D(TestCase):
    def test_origin(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(10, 0)
        ray = Ray2D(point_0, point_1)
        self.assertEqual(ray.origin().x(), 0)
        self.assertEqual(ray.dir().degree(), 0)
        point_0.set_x(20)
        self.assertEqual(ray.origin().x(), 0)
        self.assertEqual(ray.dir().degree(), 0)

    def test_dir(self):
        point_0 = Vector2D(10, 0)
        point_1 = Vector2D(0, 0)
        ray_0 = Ray2D(point_0, point_1)
        self.assertEqual(ray_0.origin().x(), 10)
        self.assertEqual(ray_0.dir().degree(), 180)

    def test_copy(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(10, 0)
        ray_0 = Ray2D(point_0, point_1)
        ray_1 = ray_0.copy()
        ray_1.origin_().set_x(15)
        self.assertEqual(ray_0.origin().x(), 0)

    def test_line(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(10, 10)
        ray_0 = Ray2D(point_0, point_1)
        line = ray_0.line()
        self.assertTrue(abs(line.b() + line.a()) < 0.1)
        self.assertEqual(line.c(), 0)

    def test_in_right_dir(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(10, 10)
        point_2 = Vector2D(1, 10)
        point_3 = Vector2D(15,13)
        ray_0 = Ray2D(point_0, point_1)
        self.assertTrue(ray_0.in_right_dir(point_1))
        self.assertFalse(ray_0.in_right_dir(point_2))
        self.assertTrue(ray_0.in_right_dir(point_3))
        self.assertFalse(ray_0.in_right_dir(point_3,2.0))

    def test_intersection(self):
        point_0 = Vector2D(-10, 0)
        point_1 = Vector2D(0, 0)
        ray_0 = Ray2D(point_0, point_1)
        point_2 = Vector2D(1, 10)
        point_3 = Vector2D(1, 5)
        ray_1 = Ray2D(point_2, point_3)
        i = ray_0.intersection(ray_1)
        self.assertTrue(abs(i.x() - 1) < 0.001)
        self.assertTrue(abs(i.y()) < 0.001)



""" size_2d.py file
    Size2D: Clase name
    Class attributes: _length, _width

"""
from __future__ import annotations
import math

from pyrusgeom.math_values import EPSILON

class Size2D:
    """ handling size of rectangle in SS2D
    Attributes:
        _length: a non-negative float for X range
        _width: a non-negative float for Y range
    """
    def __init__(self, length: float = 0.0, width: float = 0.0) -> None:
        """This is the class init function for Size2D.

        Defualt:
            create a zero area with length = 0.0 and width = 0.0
        OR:
            bulid a Size2D
        Args:
            length (float, optional): X range. Defaults to 0.0.
            width (float, optional): Y range. Defaults to 0.0.
        """
        self._length = math.fabs(length)
        self._width = math.fabs(width)

    def assign(self, length: float, width: float) -> None:
        """assign range directly.

        Args:
            length (float): X range
            width (float): Y range
        """
        self._length = math.fabs(length)
        self._width = math.fabs(width)

    def set_length(self, length: float) -> None:
        """set X range

        Args:
            length (float): X range
        """
        self._length = math.fabs(length)

    def set_width(self, width: float) -> None:
        """set Y range

        Args:
            width (float): Y range
        """
        self._width = math.fabs(width)

    def length(self) -> float:
        """get the value of X range

        Returns:
            float: value of X range
        """
        return self._length

    def width(self) -> float:
        """get the value of Y range

        Returns:
            float: value of Y range
        """
        return self._width

    def diagonal(self) -> float:
        """get the length of diagonal line

        Returns:
            float: length of diagonal line
        """
        return math.sqrt(self._length * self._length + self._width * self._width)

    def is_valid(self) -> float:
        """check if size is valid or not.

        Returns:
            float: True if the area of self rectangle is not 0. else False.
        """
        return self._length > 0.0 and self._width > 0.0

    def __eq__(self, other: Size2D) -> bool:
        """operator == for Size2D

        Args:
            other (Size2D): right hand side argument

        Returns:
            bool: true if equal or difference is less than EPSILON. else false
        """
        return math.fabs(self._length - other.length()) < EPSILON and (
            math.fabs(self._width - other.width()) < EPSILON)

    def __repr__(self) -> str:
        """represent Size2D as a string

        Returns:
            str: Size2D's _length and _width as string
        """
        return f"[len:{self._length},wid:{self._width}]"



""" region_2d.py file
    Region2D: class name
    Father Class
"""
from pyrusgeom.vector_2d import Vector2D

class Region2D:
    """ handling Regions in SS2D
        father class for other classes
    """
    def __init__(self):
        """accessible only from derived classes
        """

    def area(self) -> float:
        """get the area of this region

        Returns:
            float: value of the area
        """

    def contains(self, point: Vector2D) -> bool:
        """check if this region contains 'point'.

        Args:
            point (Vector2D): considered point

        Returns:
            bool: true or false
        """



""" polygon_2d.py file
    Polygon2D: class name
"""
from __future__ import annotations
import math

from pyrusgeom.rect_2d import Rect2D
from pyrusgeom.region_2d import Region2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.size_2d import Size2D
from pyrusgeom.line_2d import Line2D
from pyrusgeom.segment_2d import Segment2D


class XLessEqual:
    """cmp key class for point.x less or equal than threshold
    """
    def __init__(self, thr:float):
        self._thr = thr

    def __call__(self, point: Vector2D) -> bool:
        """operator <=

        Args:
            point (Vector2D): point to check

        Returns:
            bool: return true if point x is less or equal than threshold
        """
        return point.x() <= self._thr

    def thr(self) -> float:
        """return the threshold

        Returns:
            float: return key threshold
        """
        return self._thr


class XMoreEqual:
    """cmp key class for point.x more or equal than threshold
    """
    def __init__(self, thr:float):
        self._thr = thr

    def __call__(self, point: Vector2D) -> bool:
        """operator >=

        Args:
            point (Vector2D): point to check

        Returns:
            bool: return true if point x is more or equal than threshold
        """
        return point.x() >= self._thr

    def thr(self) -> float:
        """return the threshold

        Returns:
            float: return key threshold
        """
        return self._thr


class YLessEqual:
    """cmp key class for point.y less or equal than threshold
    """
    def __init__(self, thr:float):
        self._thr = thr

    def __call__(self, point: Vector2D) -> bool:
        """operator <=

        Args:
            point (Vector2D): point to check

        Returns:
            bool: return true if point y is less or equal than threshold
        """
        return point.y() <= self._thr

    def thr(self) -> float:
        """return the threshold

        Returns:
            float: return key threshold
        """
        return self._thr


class YMoreEqual:
    """cmp key class for point.y more or equal than threshold
    """
    def __init__(self, thr:float):
        self._thr = thr

    def __call__(self, point: Vector2D) -> bool:
        """operator >=

        Args:
            point (Vector2D): point to check

        Returns:
            bool: return true if point y is more or equal than threshold
        """
        return point.y() >= self._thr

    def thr(self) -> float:
        """return the threshold

        Returns:
            float: return key threshold
        """
        return self._thr


class Polygon2D(Region2D):
    """ handling polygons in SS2D

    Args:
        Region2D (mother class): each polygon is a region

    Attributes:
        _vertices : a list of vectors
    """

    def __init__(self, *args):
        """This is the class init function and creates the polygon.

            Defualt:
                create an empty polygon with one point (0,0).
            OR
                create a polygon with given points
            Args:
                none: for default empty polygon
                one:
                    list: array of input points
                else:
                    bunch of vector2D

        """
        super().__init__()
        if len(args) == 0:
            self._vertices = [Vector2D()]
        elif len(args[0]) > 0:
            self._vertices = args[0].copy()
        else:
            self._vertices = args # TODO:Check this for object init change

    def clear(self) -> None:
        """clear all data.
        """
        self._vertices = [Vector2D()]

    def assign(self, points:list[Vector2D]) -> Polygon2D:
        """set polygon with given points and returns a reference to itself
        Args:
            points(list) : points to assign
        Returns:
            Polygon2D: rturn itself
        """
        if len(points) > 0:
            self._vertices = points.copy()
        return self

    def add_vertex(self, point: Vector2D) -> None:
        """append point to polygon

        Args:
            point(Vector2d): point to add
        """
        self._vertices.append(point)

    def vertices(self) -> list[Vector2D]:
        """get a copy list from the vertex container

        Returns:
            list: a list with the vertex container values
        """
        vertices_cp = self._vertices.copy()
        return vertices_cp

    def vertices_(self) -> list:
        """ get the reference and og list to the vertex container

        Returns:
            list: a reference to the vertex container
        """
        return self._vertices

    def get_bounding_box(self) -> Rect2D:
        """get bounding box of this polygon

        Returns:
            bounding box of this polygon
        """
        if len(self._vertices) == 0:
            return Rect2D()
        x_min = float("inf")
        x_max = - float("inf")
        y_min = float("inf")
        y_max = - float("inf")
        for point in self._vertices:
            if point.x() > x_max:
                x_max = point.x()

            if point.x() < x_min:
                x_min = point.x()

            if point.y() > y_max:
                y_max = point.y()

            if point.y() < y_min:
                y_min = point.y()
        return Rect2D(Vector2D(x_min, y_min), Size2D(x_max - x_min, y_max - y_min))

    def contains(self, point: Vector2D, allow_on_segment:bool=True) :#-> bool:
        # TODO : how to OverLoad in python
        """check if given point is in this polygon or not

        Args:
            point(Vector2D): point to check
            allow_on_segment(bool): allows points on the outline

        Returns:
            bool: True if point is in this polygon (or on the line)
        """

        if len(self._vertices) <= 0:
            return False
        if len(self._vertices) == 1:
            return allow_on_segment and (self._vertices[0] == point)

        bounding_box = self.get_bounding_box()
        if not bounding_box.contains(point):
            return False

        # make virtual half line
        line = Segment2D(point, Vector2D(point.x() + ((
            bounding_box.right() - bounding_box.left() + bounding_box.bottom() - bounding_box.top())
            + (self._vertices[0] - point).r()) * 3.0, point.y()))

        # check intersection with all segments
        inside = False
        min_line_x = bounding_box.right() + 1.0
        for i in range(len(self._vertices)):
            p1_index = i + 1

            if p1_index >= len(self._vertices):
                p1_index = 0

            p_0 = self._vertices[i]
            p_1 = self._vertices[p1_index]

            if not allow_on_segment and Segment2D(p_0, p_1).on_segment(point):
                return False

            if allow_on_segment and point == p_0:
                return True

            if line.exist_intersection(Segment2D(p_0, p_1)):
                if p_0.y() == point.y() or p_1.y() == point.y():
                    if p_0.y() == point.y():
                        if p_0.x() < min_line_x:
                            min_line_x = p_0.x()

                    if p_1.y() == point.y():
                        if p_1.x() < min_line_x:
                            min_line_x = p_1.x()

                    if p_0.y() == p_1.y() or p_0.y() < point.y() or p_1.y() < point.y():
                        continue
                inside = not inside # inside * -2 #

        return inside

    def bounding_box_center(self) -> Vector2D:
        """get center of bounding box of this polygon

        Returns:
            Vector2D: center of bounding box of this polygon
        """
        return self.get_bounding_box().center()

    def dist(self, point: Vector2D, check_as_plane=True) -> float:
        """get minimum distance between this polygon and point

        Args:
            point(Vector2d): point to check
            check_as_plane(bool): if this parameter is true, the polygon
            counts as a plane polygon otherwise as a polyline polygon.

        Returns:
            float: minimum distance between this polygon and point
        """
        size = len(self._vertices)

        if size == 1:
            return (self._vertices[0] - point).r()

        if check_as_plane and self.contains(point):
            return 0.0

        min_dist = float('inf')
        for i in range(size - 1):

            seg = Segment2D(self._vertices[i],
                            self._vertices[i + 1])

            point_dist = seg.dist(point)

            if point_dist < min_dist:
                min_dist = point_dist

        if size >= 3:
            seg = Segment2D(self._vertices[size - 1], self._vertices[0])

            point_dist = seg.dist(point)

            if point_dist < min_dist:
                min_dist = point_dist

        return min_dist


    def area(self) -> float:
        """get area of this polygon

        Returns:
            float: value of area with sign.
        """
        return math.fabs(self.double_signed_area() * 0.5)

    def double_signed_area(self) -> float:
        """calculate doubled signed area value

        If vertices are placed counterclockwise order, positive number.
        If vertices are placed clockwise order, negative number.
        Otherwise, 0.

        Returns:
            float: value of doubled signed area.
        """
        size = len(self._vertices)
        ds_area_value = 0.0

        if size < 3:
            return ds_area_value

        for i in range(size):
            n_count = i + 1
            if n_count == size:
                n_count = 0
            ds_area_value += (self._vertices[i].x() * self._vertices[n_count].y()
                              - self._vertices[n_count].x() * self._vertices[i].y())

        return ds_area_value

    def is_counter_clockwise(self) -> bool:
        """check vertexes of self polygon is placed counterclockwise ot not

        Returns:
            bool: True if counterclockwise. else false
        """
        return self.double_signed_area() > 0.0

    def is_clockwise(self) -> bool:
        """check vertexes of self polygon is placed clockwise ot not

        Returns:
            bool: True if clockwise. else false
        """
        return self.double_signed_area() < 0.0

    def get_rectangle_clipped_polygon(self, rect: Rect2D) -> Polygon2D:
        """get a polygon clipped and cropped by a rectangle

        if polygon is divided into more by edges of rectangle, each
        separated polygon is going to connect to the frist polygon.

        Args:
            rect (Rect2D): rectangle for clipping

        Returns:
            Polygon2D: a united polygon
        """
        if len(self._vertices) == 0:
            return Polygon2D()

        points = self._vertices
        clipped_p_1 = []
        clipped_p_2 = []
        clipped_p_3 = []
        clipped_p_4 = []
        clipped_p_1 = Polygon2D.get_line_clipped_polygon(XLessEqual(rect.right()),
                                                points,
                                                Line2D(Vector2D(rect.right(), 0.0), AngleDeg(90.0)))

        clipped_p_2 = Polygon2D.get_line_clipped_polygon(YLessEqual(rect.bottom()),
                                                clipped_p_1,
                                                Line2D(Vector2D(0.0, rect.top()), AngleDeg(0.0)))

        clipped_p_3 = Polygon2D.get_line_clipped_polygon(XMoreEqual(rect.left()),
                                                clipped_p_2,
                                                Line2D(Vector2D(rect.left(), 0.0), AngleDeg(90.0)))

        clipped_p_4 = Polygon2D.get_line_clipped_polygon(YMoreEqual(rect.top()),
                                                clipped_p_3,
                                                Line2D(Vector2D(0.0, rect.top()), AngleDeg(0.0)))

        return Polygon2D(clipped_p_4)

    @staticmethod
    def get_line_clipped_polygon(in_region, points:list[Vector2D], line:Line2D) -> list[Vector2D]:
        """get points clipped by a line

        Args:
            in_geion(side_class): cmp key class - XLessEqual,YLessEqual,XMoreEqual,YMoreEqual
            points(list[Vector2D]): list of points
            line(Line2D): line for clipping

        Returns:
            list[Vector2D]: a list contains the final points
        """
        new_points = []
        in_rectangle = []
        for point in points:
            in_rectangle.append(in_region(point))
        for i in range(len(points)):
            index_0 = i
            index_1 = i + 1
            if index_1 >= len(points):
                index_1 = 0

            p_0 = points[index_0]
            p_1 = points[index_1]

            if in_rectangle[index_0]:
                if in_rectangle[index_1]:
                    new_points.append(p_1)
                else:
                    cuted = line.intersection(Line2D(p_0, p_1))

                    if not cuted.is_valid():
                        return new_points
                    new_points.append(cuted)
            else:
                if in_rectangle[index_1]:
                    cuted = line.intersection(Line2D(p_0, p_1))

                    if not cuted.is_valid():
                        return new_points

                    new_points.append(cuted)
                    new_points.append(p_1)
        return new_points

    def __repr__(self):
        """represent the polygon as a string

        Returns:
            str: contains _vertices
        """
        return f"({self._vertices})"


# def test():
#     p = Polygon2D([Vector2D(0, 0), Vector2D(0, 4), Vector2D(4, 4), Vector2D(4, 0)])
#     v = [Vector2D(2, 2), Vector2D(5, 5)]
#     print(p)
#     print(p.is_clockwise(), p.is_counter_clockwise())
#     print(p.double_signed_area(), p.area())
#     print(p.get_bounding_box())
#     print(p.contains(v[0]), p.contains(v[1]))



""" circle_2d.py file
    Circle2D: class name
    Class attributes: _center, _radius, _is_valid
    TODO: add eq hash

"""
from __future__ import annotations
from typing import Union
import math

from pyrusgeom.segment_2d import Segment2D
from pyrusgeom.ray_2d import Ray2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.line_2d import Line2D
from pyrusgeom.triangle_2d import Triangle2D
from pyrusgeom.math_values import PI, EPSILON


class Circle2D:
    """ handling circles in SS2D
    Attributes:
        _center: a Vector2D for center point
        _radius: a non-negetive float for radius value
        _is_valid: a boolean for validation
    """

    def __init__(self, center: Vector2D = Vector2D(), radius: float = 0.0):
        """This is the class init function for Circle2D.

        Defualt:
            create a zero area circle at (0,0) which has zero radius
        OR:
            bulid a circle2d
        Args:
            center (Vector2D, optional): center point. Defaults to Vector2D().
            radius (float, optional): a non-negetive radius value. Defaults to 0.0.
        """
        self._center: Vector2D = Vector2D(center)
        self._radius: float = radius
        if radius < 0.0:
            self._radius = 0.0
        self._is_valid = True

    def assign(self, center: Vector2D, radius: float):
        """set a new circle for Circle2D

        Args:
            center (Vector2D): center point
            radius (float): radius value
        """
        self._center = Vector2D(center)
        self._radius = radius
        if radius < 0.0:
            self._radius = 0.0
        self._is_valid = True

    def area(self) -> float:
        """get the area value of this circle

        Returns:
            float: value of area of circle
        """
        return PI * self._radius * self._radius

    def contains(self, point: Vector2D) -> bool:
        """check if point is within this circle

        Args:
            point (Vector2D): considered point

        Returns:
            bool: True if point is contained by self circle
        """
        return self._center.dist2(point) < self._radius * self._radius

    def center(self) -> Vector2D:
        """get the center point copy

        Returns:
            Vector2D: center point coordinate value
        """
        return Vector2D(self._center)

    def center_(self) -> Vector2D:
        """get the center point

        Returns:
            Vector2D: center point coordinate
        """
        return self._center

    def radius(self) -> float:
        """get the radius value

        Returns:
            float: radius value
        """
        return self._radius

    def intersection(self, other: Union[Line2D, Ray2D, Segment2D, Circle2D]) -> list[Vector2D]:
        """calculate the intersection with Line2D, Ray2D, Segment2D, Circle2D
        Args:
            other (Union[Line2D, Ray2D, Segment2D, Circle2D]): considered line
                    or ray or segment or circle

        Returns:
            list[Vector2D]: a list contains solutions if there is none return an empty list
        Raises:
            Exception: Input must be one of Line2D, Ray2D, Segment2D, Circle2D
        """
        if isinstance(other, Line2D):
            if math.fabs(other.a()) < EPSILON:
                if math.fabs(other.b()) < EPSILON:
                    return []

                n_sol = quadratic_f(1.0,
                                    -2.0 * self._center.x(),
                                    (math.pow(self._center.x(), 2)
                                     + math.pow(other.c() /
                                                other.b() + self._center.y(), 2)
                                     - math.pow(self._radius, 2)))

                if len(n_sol):
                    sol_tmp_1 = -other.c() / other.b()
                    sol_1 = n_sol[0]
                    sol_2 = n_sol[1]
                    sol_list = [Vector2D(sol_1, sol_tmp_1),
                                Vector2D(sol_2, sol_tmp_1)]
                    if sol_list[0].equals_weakly(sol_list[1]):
                        del sol_list[1]
                else:
                    sol_list = []
                return sol_list

            b_a = other.b() / other.a()
            c_a = other.c() / other.a()

            line_a = 1.0 + b_a * b_a
            line_b = 2.0 * (-self._center.y() + (c_a + self._center.x()) * b_a)
            line_c = (c_a + self._center.x()) ** 2 + \
                (self._center.y()) ** 2 - self._radius ** 2

            n_sol = quadratic_f(line_a, line_b, line_c)
            if len(n_sol):
                sol_tmp_1 = n_sol[0]
                sol_tmp_2 = n_sol[1]
                sol_list = [Vector2D(other.get_x(sol_tmp_1), sol_tmp_1),
                            Vector2D(other.get_x(sol_tmp_2), sol_tmp_2)]
                if sol_list[0].equals_weakly(sol_list[1]):
                    del sol_list[1]
            else:
                sol_list = []

            return sol_list
        if isinstance(other, Ray2D):
            line_tmp = Line2D(other.origin(), other.dir())
            sol_list = self.intersection(line_tmp)
            if len(sol_list) > 1 and not other.in_right_dir(sol_list[1], 1.0):
                del sol_list[1]

            if len(sol_list) > 0 and not other.in_right_dir(sol_list[0], 1.0):
                del sol_list[0]

            return sol_list

        if isinstance(other, Segment2D):
            line = other.line()
            sol_list = self.intersection(line)
            if len(sol_list) > 1 and not other.contains(sol_list[1]):
                del sol_list[1]

            if len(sol_list) > 0 and not other.contains(sol_list[0]):
                del sol_list[0]

            return sol_list

        if isinstance(other, Circle2D):
            rel_x = other.center().x() - self._center.x()
            rel_y = other.center().y() - self._center.y()
            center_dist2 = rel_x * rel_x + rel_y * rel_y
            center_dist = math.sqrt(center_dist2)
            if (center_dist < math.fabs(self._radius - other.radius()) or
                    self._radius + other.radius() < center_dist):
                return []

            line = Line2D(-2.0 * rel_x, -2.0 * rel_y,
                          other.center().r2() - other.radius() * other.radius()
                          - self._center.r2() + self._radius * self._radius)
            sol_list = self.intersection(line)
            return sol_list

        raise Exception(
            'The input should be one of Line2D, Ray2D, Segment2D, Circle2D')

    def intersection_with_line(self, line: Line2D) -> list[Vector2D]:
        """calculate the intersection with Line2D

        Args:
            line (Line2D): considered line

        Returns:
            list[Vector2D]: a list contains solutions if there is none return an empty list
        """
        return self.intersection(line)

    def intersection_with_ray(self, ray: Ray2D) -> list[Vector2D]:
        """calculate the intersection with Ray2D

        Args:
            line (Ray2D): considered ray

        Returns:
            list[Vector2D]: a list contains solutions if there is none return an empty list
        """
        return self.intersection(ray)

    def intersection_with_segment(self, segment: Segment2D) -> list[Vector2D]:
        """calculate the intersection with Segment2D

        Args:
            line (Segment2D): considered segment

        Returns:
            list[Vector2D]: a list contains solutions if there is none return an empty list
        """
        return self.intersection(segment)

    def intersection_with_circle(self, circle: Circle2D) -> list[Vector2D]:
        """calculate the intersection with Circle2D

        Args:
            line (Circle2D): considered circle

        Returns:
            list[Vector2D]: a list contains solutions if there is none return an empty list
        """
        return self.intersection(circle)

    @staticmethod
    def circum_circle(point_0: Vector2D, point_1: Vector2D, point_2: Vector2D) -> Circle2D:
        """get the circle through three points (circum_circle of the triangle).

        Args:
            point_0 (Vector2D): triangle's 1st vertex
            point_1 (Vector2D): triangle's 2nd vertex
            point_2 (Vector2D): triangle's 3rd vertex

        Returns:
            Circle2D: coordinates of circumcenter
        """

        center = Triangle2D.tri_circumcenter(point_0, point_1, point_2)

        if not center.is_valid():
            return Circle2D()

        return Circle2D(center, center.dist(point_0))

    @staticmethod
    def circum_circle_contains(input_point: Vector2D, point_0: Vector2D,
                               point_1: Vector2D, point_2: Vector2D) -> bool:
        """check if the circum_circle contains the input point

        Args:
            input_point (Vector2D): input point
            point_0 (Vector2D): triangle's 1st vertex
            point_1 (Vector2D): triangle's 2nd vertex
            point_2 (Vector2D): triangle's 3rd vertex

        Returns:
            bool: True if circum_circle contains the point, False.
        """

        p_1_0_x = point_1.x() - point_0.x()
        p_1_0_y = point_1.y() - point_0.y()
        p_2_0_x = point_2.x() - point_0.x()
        p_2_0_y = point_2.y() - point_0.y()

        p_1_0_avg = p_1_0_x * (point_0.x() + point_1.x()) + \
            p_1_0_y * (point_0.y() + point_1.y())
        p_2_0_avg = p_2_0_x * (point_0.x() + point_2.x()) + \
            p_2_0_y * (point_0.y() + point_2.y())

        calc_area = 2.0 * (p_1_0_x * (point_2.y() - point_1.y()) -
                           p_1_0_y * (point_2.x() - point_1.x()))
        if math.fabs(calc_area) < 1.0e-10:
            return False

        center = Vector2D((p_2_0_y * p_1_0_avg - p_1_0_y * p_2_0_avg) / calc_area,
                          (p_1_0_x * p_2_0_avg - p_2_0_x * p_1_0_avg) / calc_area)
        return center.dist2(input_point) < center.dist2(point_0) - EPSILON * EPSILON

    def __repr__(self) -> str:
        """represent Circle2D as a string

        Returns:
            str: Circle2D's center and radius as string
        """
        return f"({self._center} , {self._radius})"

    def to_str(self, ostr: str) -> None:
        """add circle to the ostr

        Args:
            ostr (str): str to add to it
        """
        ostr += f'(circle {round(self.center().x(), 3)} \
            {round(self.center().y(), 3)} {round(self.radius(), 3)})'


def quadratic_f(qf_a: float, qf_b: float, qf_c: float) -> list[float]:
    """solve quadratic equation
    Args:
        qf_a (float): formula constant A
        qf_b (float): formula constant B
        qf_c (float): formula constant C

    Returns:
        list[float]: sol1 reference to the result variable, sol2 reference to the result variable
    """

    delta = qf_b * qf_b - 4.0 * qf_a * qf_c
    sol1 = 0.0
    sol2 = 0.0
    if qf_a == 0:
        raise Exception('Dvided by zero in quadratic_f')
    if math.fabs(delta) < EPSILON:
        sol1 = -qf_b / (2.0 * qf_a)
        return [sol1,sol1]
    elif delta < 0.0:
        return []
    else:
        delta = math.sqrt(delta)
        sol1 = (-qf_b + delta) / (2.0 * qf_a)
        sol2 = (-qf_b - delta) / (2.0 * qf_a)
        return [sol1, sol2]



""" Common Conset Mathmatices velues and minmax error value
    PI / EPSILON / DEG2RAD / RAD2DEG / CALC_ERROR / ERROR_VALUE
"""
PI = 3.14159265358979323846
EPSILON = 1.0e-6
DEG2RAD = PI / 180.0
RAD2DEG = 180.0 / PI
CALC_ERROR = 1.0e-9
ERROR_VALUE = 1e20



""" WildCard Import
"""
from pyrusgeom.angle_deg import *
from pyrusgeom.vector_2d import *
from pyrusgeom.line_2d import *
from pyrusgeom.ray_2d import *
from pyrusgeom.sector_2d import *
from pyrusgeom.triangle_2d import Triangle2D
from pyrusgeom.segment_2d import Segment2D
from pyrusgeom.matrix_2d import *  # not necessary but optimized for 2d
from pyrusgeom.circle_2d import *
from pyrusgeom.rect_2d import *
from pyrusgeom.polygon_2d import *



from unittest import TestCase
from pyrusgeom.circle_2d import Circle2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.line_2d import Line2D
from pyrusgeom.ray_2d import Ray2D
from pyrusgeom.segment_2d import Segment2D


class TestCircle2D(TestCase):
    def test_assign(self):
        c = Circle2D(Vector2D(0, 0), 2)
        self.assertEqual(c.radius(), 2)
        self.assertEqual(c.center().x(), 0)
        self.assertEqual(c.center().y(), 0)

    def test_area(self):
        c = Circle2D(Vector2D(0, 0), 2)
        self.assertTrue(12.56 < c.area() < 12.57)

    def test_contains(self):
        c = Circle2D(Vector2D(0, 0), 2)
        self.assertTrue(c.contains(Vector2D(1, 1)))
        self.assertFalse(c.contains(Vector2D(10, 10)))

    def test_center(self):
        pass

    def test_radius(self):
        pass

    def test_intersection(self):
        c = Circle2D(Vector2D(0, 0), 2)
        line = Line2D(Vector2D(0, 0), 0)
        sols = c.intersection(line)
        self.assertEqual(len(sols), 2)
        self.assertEqual(sols[0].x(), 2)
        self.assertEqual(sols[1].x(), -2)
        self.assertEqual(sols[0].y(), 0)
        self.assertEqual(sols[1].y(), 0)

        line = Line2D(Vector2D(0, 2), 0)
        sols = c.intersection(line)
        self.assertEqual(len(sols), 1)
        self.assertEqual(sols[0].x(), 0)
        self.assertEqual(sols[0].y(), 2)

        line = Line2D(Vector2D(0, 3), 0)
        sols = c.intersection(line)
        self.assertEqual(len(sols), 0)

        ray = Ray2D(Vector2D(0, 0), Vector2D(3, 0))
        sols = c.intersection(ray)
        self.assertEqual(len(sols), 1)
        self.assertEqual(sols[0].x(), 2)
        self.assertEqual(sols[0].y(), 0)

        ray = Ray2D(Vector2D(3, 3), Vector2D(5, 3))
        sols = c.intersection(ray)
        self.assertEqual(len(sols), 0)

        ray = Ray2D(Vector2D(-3, 0), Vector2D(3, 0))
        sols = c.intersection(ray)
        self.assertEqual(len(sols), 2)
        self.assertEqual(sols[0].x(), 2)
        self.assertEqual(sols[0].y(), 0)
        self.assertEqual(sols[1].x(), -2)
        self.assertEqual(sols[1].y(), 0)

        ray = Ray2D(Vector2D(-2, 2), Vector2D(2, 2))
        sols = c.intersection(ray)
        self.assertEqual(len(sols), 1)
        self.assertEqual(sols[0].x(), 0)
        self.assertEqual(sols[0].y(), 2)

        seg = Segment2D(Vector2D(0, 0), Vector2D(3, 0))
        sols = c.intersection(seg)
        self.assertEqual(len(sols), 1)
        self.assertEqual(sols[0].x(), 2)
        self.assertEqual(sols[0].y(), 0)

        seg = Segment2D(Vector2D(5, 5), Vector2D(8, 8))
        sols = c.intersection(seg)
        self.assertEqual(len(sols), 0)

        seg = Segment2D(Vector2D(-3, 0), Vector2D(3, 0))
        sols = c.intersection(seg)
        self.assertEqual(len(sols), 2)
        self.assertEqual(sols[0].x(), 2)
        self.assertEqual(sols[0].y(), 0)
        self.assertEqual(sols[1].x(), -2)
        self.assertEqual(sols[1].y(), 0)

        seg = Segment2D(Vector2D(-2, 2), Vector2D(2, 2))
        sols = c.intersection(seg)
        self.assertEqual(len(sols), 1)
        self.assertEqual(sols[0].x(), 0)
        self.assertEqual(sols[0].y(), 2)

        c2 = Circle2D(Vector2D(4, 0), 2)
        sols = c.intersection(c2)
        self.assertEqual(len(sols), 1)
        self.assertEqual(sols[0].x(), 2)
        self.assertEqual(sols[0].y(), 0)

        c2 = Circle2D(Vector2D(4, 0), 10)
        sols = c.intersection(c2)
        self.assertEqual(len(sols), 0)

        c2 = Circle2D(Vector2D(4, 0), 3)
        sols = c.intersection(c2)
        self.assertEqual(len(sols), 2)
        print(sols)
        self.assertEqual(sols[0].x(), 1.375)
        self.assertTrue(abs(sols[0].y() - 1.45) < 0.1)

    def test_circum_circle(self):
        v1 = Vector2D(0, 0)
        v2 = Vector2D(1, 0)
        v3 = Vector2D(0, 1)
        circle = Circle2D.circum_circle(v1, v2, v3)
        self.assertEqual(circle.center().x(), 0.5)
        self.assertEqual(circle.center().y(), 0.5)
        self.assertAlmostEqual(circle.radius(), 0.70710678)

    def test_circle_contains(self):
        v1 = Vector2D(0, 0)
        v2 = Vector2D(1, 0)
        v3 = Vector2D(0, 1)
        self.assertTrue(Circle2D.circum_circle_contains(Vector2D(0.5, 0.5), v1, v2, v3))
        self.assertFalse(Circle2D.circum_circle_contains(Vector2D(2, 2), v1, v2, v3))



""" vector_2d.py
    Vector2D: class name
    Clase attributes: _x, _y, _is_valid
"""
from __future__ import annotations
from typing import Union
import math

from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.math_values import EPSILON, DEG2RAD


class Vector2D:
    """ handling vectors and points in SS2D
    Attributes:
        _x: x-coordinate
        _y: y-coordinate
        _is_valid: a boolean for validation
    """

    def __init__(self, *args, **kwargs) -> None:
        """brief default constructor : create a Vector2D with XY value directly.

        Defualt:
            create a Vector2D at (0,0)
        Or
            create a Vector2D at (X,Y)

        Args:
            Two:
                float, float:
                    x: a float for x
                    y: a float for y
            One:
                Vector2D:
                    point: a Vector2D to bulid from.
            None:
                create a Vector2D at (0,0)

        Kwargs:
            x: a float for x
            y: a float for y
            r (Union[int, float]): vector's radius
            a (Union[int, float, AngleDeg]): vector's angle

        Raises:
            Exception: The input should be a Vector2D or two numbers or no input
        """
        self._is_valid = True
        if len(kwargs.keys()) == 0:
            if len(args) == 2 and isinstance(args[0],
                                            (int, float)) and isinstance(args[1], (int, float)):
                self._x = args[0]
                self._y = args[1]
            elif len(args) == 1 and isinstance(args[0], Vector2D):
                self._x = args[0].x()
                self._y = args[0].y()
            elif len(args) == 0:
                self._x = 0
                self._y = 0
            else:
                raise Exception('The input should be a Vector2D or two numbers')
        elif len(kwargs) == 2:
            if 'x' in kwargs and 'y' in kwargs:
                self._x = kwargs['x']
                self._y = kwargs['y']
            elif 'r' in kwargs and 'a' in kwargs:
                self.set_polar(kwargs['r'], kwargs['a'])
            else:
                raise Exception('The input should be a Vector2D or two numbers')
        else:
            raise Exception('The input should be a Vector2D or two numbers')


    def x(self) -> float:
        """accessor to x

        Returns:
            float: X coordinate
        """
        return self._x

    def y(self) -> float:
        """accessor to y

        Returns:
            float: Y coordinate
        """
        return self._y

    def get_x(self) -> float:
        """accessor to x

        Returns:
            float: X coordinate
        """
        return self._x

    def get_y(self) -> float:
        """accessor to y

        Returns:
            float: Y coordinate
        """
        return self._y

    def copy(self) -> Vector2D:
        """returns copy of this Vector2D

        Returns:
           Vector2D: Copy of this Vector 2D
        """
        return Vector2D(self._x, self._y)

    def get_copy(self) -> Vector2D:
        """returns copy of this Vector2D

        Returns:
           Vector2D: Copy of this Vector 2D
        """
        return self.copy()

    def assign(self, p_x: float, p_y: float) -> Vector2D:
        """assign XY value directly.

        Args:
            p_x (float): assigned x value
            p_y (float): assigned y value

        Returns:
            Vector2D: reference to itself
        """
        self._x = p_x
        self._y = p_y
        return self

    def set_x(self, p_x: Union[int, float]) -> None:
        """set x of this object to new x.

        Args:
            p_x (Union[int, float]): assigned x value
        """
        self._x = p_x

    def set_y(self, p_y: Union[int, float]) -> None:
        """set y of this object to new y.

        Args:
            p_y (Union[int, float]): assigned y value
        """
        self._y = p_y

    def set_x_y(self, p_x: Union[int, float], p_y: Union[int, float]) -> None:
        """brief set x and y of this object to new x and y.

        Args:
            p_x (Union[int, float]): assigned x value
            p_y (Union[int, float]): assigned y value
        """
        self._x = p_x
        self._y = p_y

    def set_vector(self, other: Vector2D) -> None:
        """set x and y of this object to the x and y of the new object.

        Args:
            other (Vector2D): A vector 2D
        """
        self._x = other.x()
        self._y = other.y()
        self._is_valid = other.is_valid()

    def set_polar(self, radius: Union[int, float], angle: Union[int, float, AngleDeg]) -> None:
        """assign XY value from POLAR value.

        Args:
            radius (Union[int, float]): vector's radius
            angle (Union[int, float, AngleDeg]): vector's angle
        """
        if not isinstance(angle, AngleDeg):
            angle = AngleDeg(angle)
        self._x = radius * angle.cos()
        self._y = radius * angle.sin()

    def validate(self) -> None:
        """validate this object

        _is_valid = True
        """
        self._is_valid = True

    def invalidate(self) -> None:
        """invalidate this object

        _is_valid = False
        """
        self._is_valid = False

    def is_valid(self) -> bool:
        """check this object is valid

        Returns:
            bool: True if valid. else False
        """
        return self._is_valid

    def r2(self) -> float:
        """get the squared length of vector.
        Returns:
            float: squared length value
        """
        return self._x * self._x + self._y * self._y

    def r(self) -> float:
        """get the length of vector.

        Returns:
            float: length value
        """
        return math.sqrt(self.r2())

    def length(self) -> float:
        """get the length of vector.

        this method is equivalent to r().

        Returns:
            float: length value
        """
        return self.r()

    def length2(self):
        """get the squared length of vector.

        this method is equivalent to r2().

        Returns:
            float: squared length value
        """
        return self.r2()

    def th(self) -> AngleDeg:
        """get the angle of vector.

        Returns:
            AngleDeg: the angle
        """
        return AngleDeg(AngleDeg.atan2_deg(self._y, self._x))

    def dir(self) -> AngleDeg:
        """get the angle of vector.

        this method is equivalent to th().

        Returns:
            AngleDeg: the angle
        """
        return self.th()

    def abs(self) -> Vector2D:
        """get new vector that XY values were set to absolute value.

        Returns:
            Vector2D:a new vector with absolute values.
        """
        return Vector2D(abs(self._x), abs(self._y))

    def abs_x(self) -> float:
        """get absolute x value

        Returns:
            float: absolute x value
        """
        return math.fabs(self._x)

    def abs_y(self):
        """get absolute y value

        Returns:
            float: absolute y value
        """
        return math.fabs(self._y)

    def add(self, *args) -> None:
        """adds given x and y to this Vector2D

        Args:
            Two:
                float, float:
                    x: a float for x
                    y: a float for y
            One:
                Vector2D:
                    point: a Vector2D to add from.

        Raises:
            Exception: The input should be a Vector2D or two numbers
        """
        if len(args) == 1 and isinstance(args[0], Vector2D):
            self._x += args[0].x()
            self._y += args[0].y()
        elif len(args) == 2:
            self._x += args[0]
            self._y += args[1]
        else:
            raise Exception("The input should be a Vector2D or two numbers")

    def add_x(self, p_x: Union[int, float]) -> None:
        """adds p_x to this Vector's X

        Args:
            p_x (Union[int, float]): a float for x
        """
        self._x += p_x

    def add_y(self, p_y: Union[int, float]) -> None:
        """adds p_y to this Vector's Y

        Args:
            p_y (Union[int, float]): a float for y
        """
        self._y += p_y

    def sub(self, *args) -> None:
        """subtracts x and y from this vector 2d

        Args:
            Two:
                float, float:
                    x: a float for x
                    y: a float for y
            One:
                Vector2D:
                    point: a Vector2D to add from.

        Raises:
            Exception: The input should be a Vector2D or two numbers
        """
        if len(args) == 1:
            self._x -= args[0].x()
            self._y -= args[0].y()
        elif len(args) == 2:
            self._x -= args[0]
            self._y -= args[1]
        else:
            raise Exception("The input should be a Vector2D or two numbers")

    def sub_x(self, p_x: Union[int, float]) -> None:
        """subtracts p_x from this Vector's X

        Args:
            p_x (Union[int, float]): a float for x
        """
        self._x -= p_x

    def sub_y(self, p_y: Union[int, float]) -> None:
        """subtracts p_y from this Vector's Y

        Args:
            p_y (Union[int, float]): a float for y
        """
        self._y -= p_y

    def scale(self, scalar: Union[int, float]) -> None:
        """scale this vector

        Args:
            scalar (Union[int, float]): scaling factor
        """
        self._x *= scalar
        self._y *= scalar

    def dist2(self, other: Vector2D) -> float:
        """get the squared distance from this to 'other'.

        Args:
            other (Vector2D): target point

        Returns:
            float: squared distance to 'other'
        """
        return math.pow(self._x - other.x(), 2) + math.pow(self._y - other.y(), 2)

    def dist(self, other: Vector2D) -> float:
        """get the distance from this Vector2D to the 'other' Vector2D.

        Args:
            other (Vector2D): target point

        Returns:
            float: distance to 'other'
        """
        return math.sqrt(self.dist2(other))

    def reverse(self) -> Vector2D:
        """reverse this vector components

        Returns:
            Vector2D: self
        """
        self._x *= (-1.0)
        self._y *= (-1.0)
        return self

    def reverse_vector(self) -> Vector2D:
        """get reversed vector.

        Returns:
            Vector2D: new reversed vector object
        """
        return Vector2D(self._x, self._y).reverse()

    def set_length(self, length: Union[int, float]) -> None:
        """set vector length to 'length'

        Args:
            length (Union[int, float]): new length to set
        """
        mag = self.r()
        if mag > EPSILON:
            self.scale(length / mag)

    def set_length_vector(self, length: Union[int, float]) -> Vector2D:
        """create a new vector from this vertor with the length is set to 'length'

        Args:
            length (Union[int, float]): new length

        Returns:
            Vector2D: new vector that the length is set to 'length'
        """
        new_vector = Vector2D(self._x, self._y)
        new_vector.set_length(length)
        return new_vector

    def normalize(self) -> None:
        """normalize vector.

        length is set to 1.0.
        """
        self.set_length(1)

    def normalize_vector(self) -> Vector2D:
        """get new normalized vector that the length is set to 1.0 with the same angle as this
        Returns:
            Vector2D: new normalized vector
        """
        vector = Vector2D(self._x, self._y)
        vector.set_length(1)
        return vector

    def inner_product(self, point: Vector2D) -> float:
        """get inner(dot) product with 'point'.

        |this| * |point| * (this - point).th().cos()

        Args:
            point (Vector2D): target vector

        Returns:
            float: value of inner product
        """
        return self._x * point.x() + self._y * point.y()

    def outer_product(self, point: Vector2D) -> float:
        """get virtual outer(cross) product with 'point'.

        xn = self.y * point.z - self.z * point.y;
        yn = self.z * point.x - self.x * point.z;
        zn = self.x * point.y - self.y * point.x;

        |this| * |point| * (this - point).th().sin()

        Args:
            point (Vector2D): target vector

        Returns:
            float: value of outer product
        """
        return self._x * point.y() - self._y * point.x()

    def equals(self, other: Vector2D) -> bool:
        """check if this vector is strictly same as given vector.

        Args:
            other (Vector2D): compared vector

        Returns:
            bool: true if strictly same, otherwise false.
        """
        return self._x == other.x() and self._y == other.y()

    def equals_weakly(self, other: Vector2D) -> bool:
        """check if this vector is weakly same as given vector.

        Args:
            other (Vector2D): compared vector.

        Returns:
            bool: true if weakly same, otherwise false.
        """
        return math.fabs(self._x - other.x()) < EPSILON and math.fabs(self._y - other.y()) < EPSILON

    def rotate(self, deg: Union[int, float, AngleDeg]) -> Vector2D:
        """rotate this vector with 'deg'

        Args:
            deg (Union[int, float, AngleDeg]): rotated angle.

        Returns:
            Vector2D: rotated vector
        """
        if isinstance(deg, AngleDeg):
            deg = deg.degree()
        cos_tmp = math.cos(deg * DEG2RAD)
        sin_tmp = math.sin(deg * DEG2RAD)
        return self.assign(self._x * cos_tmp - self._y * sin_tmp,
                           self._x * sin_tmp + self._y * cos_tmp)

    def rotated_vector(self, deg: Union[int, float, AngleDeg]) -> Vector2D:
        """get new vector that is rotated by 'deg'.

        Args:
            deg (Union[int, float, AngleDeg]): rotated angle.

        Returns:
            Vector2D: new rotated vector by 'deg'
        """
        return Vector2D(self._x, self._y).rotate(deg)

    def set_dir(self, direction: Union[int, float, AngleDeg]) -> None:
        """set vector's angle to 'angle'

        Args:
            direction (Union[int, float, AngleDeg]): new angle to be set
        """
        if not isinstance(direction, AngleDeg):
            direction = AngleDeg(direction)
        radius = self.r()
        self._x = radius * direction.cos()
        self._y = radius * direction.sin()

    #  __ operator section __

    __hash__ = None

    def __eq__(self, other: Vector2D) -> bool:
        return isinstance(other, Vector2D) and self._x == other.x() and self._y == other.y()

    def __add__(self, other: Vector2D) -> Vector2D:
        return Vector2D(self._x + other.x(), self._y + other.y())

    def __sub__(self, other: Vector2D) -> Vector2D:
        return Vector2D(self._x - other._x, self._y - other._y)

    def __truediv__(self, other: Union[int, float]) -> Vector2D:
        return Vector2D(self._x / other, self._y / other)

    def __mul__(self, other: Union[int, float]) -> Vector2D:
        return Vector2D(self._x * other, self._y * other)

    def __iadd__(self, other: Vector2D) -> Vector2D:
        self._x += other.x()
        self._y += other.y()
        return self

    def __isub__(self, other: Vector2D) -> Vector2D:
        self._x -= other.x()
        self._y -= other.y()
        return self

    def __imul__(self, other: Union[int, float]) -> Vector2D:
        self._x *= other
        self._y *= other
        return self

    def __idiv__(self, other: Union[int, float]) -> Vector2D:
        self._x /= other
        self._y /= other
        return self

    def __repr__(self):
        return f"({self._x},{self._y})"

    @staticmethod
    def invalid() -> Vector2D:
        """make an invalid vector2D

        Returns:
            Vector2D: invalid vector
        """
        vec_invalid = Vector2D()
        vec_invalid.invalidate()
        return vec_invalid

    @staticmethod
    def from_polar(mag: Union[int, float], theta: Union[int, float, AngleDeg]) -> Vector2D:
        """get new Vector created by POLAR value.

        Args:
            mag (Union[int, float]): length of vector
            theta (Union[int, float, AngleDeg]): angle of vector

        Returns:
            Vector2D: new vector object
        """
        if not isinstance(theta, AngleDeg):
            theta = AngleDeg(theta)
        return Vector2D(mag * theta.cos(), mag * theta.sin())

    @staticmethod
    def polar2vector(radius: Union[int, float], direction: Union[int, float, AngleDeg]) -> Vector2D:
        """get new Vector created by POLAR value.

        Args:
            radius (Union[int, float]): length of vector
            direction (Union[int, float, AngleDeg]): angle of vector

        Returns:
            Vector2D: new vector object
        """
        return Vector2D.from_polar(radius, direction)

    @staticmethod
    def inner_product_static(vec_1: Vector2D, vec_2: Vector2D) -> float:
        """get inner(dot) product for vec_1 and vec_2.

        Args:
            vec_1 (Vector2D): 1st vector
            vec_2 (Vector2D): 2nd vector

        Returns:
            float: value of inner product
        """
        return vec_1.inner_product(vec_2)

    @staticmethod
    def outer_product_static(vec_1: Vector2D, vec_2: Vector2D) -> float:
        """get outer(cross) product for vec_1 and vec_2.

        Args:
            vec_1 (Vector2D): 1st vector
            vec_2 (Vector2D): 2nd vector

        Returns:
            float: value of outer product
        """
        return vec_1.outer_product(vec_2)



from unittest import TestCase
from pyrusgeom.segment_2d import Segment2D
from pyrusgeom.vector_2d import Vector2D


class TestSegment2D(TestCase):
    def test_general(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(10, 2)
        seg_0 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.origin().x(), 0)
        self.assertEqual(seg_0.origin().y(), 1)
        self.assertEqual(seg_0.terminal().x(), 10)
        self.assertEqual(seg_0.terminal().y(), 2)
        seg_0 = Segment2D(0, 1, 10, 2)
        self.assertEqual(seg_0.origin().x(), 0)
        self.assertEqual(seg_0.origin().y(), 1)
        self.assertEqual(seg_0.terminal().x(), 10)
        self.assertEqual(seg_0.terminal().y(), 2)
        seg_0 = Segment2D(point_0, 10, 0)
        self.assertEqual(seg_0.origin().x(), 0)
        self.assertEqual(seg_0.origin().y(), 1)
        self.assertEqual(seg_0.terminal().x(), 10)
        self.assertEqual(seg_0.terminal().y(), 1)

    def test_assign(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(10, 2)
        seg_0 = Segment2D()
        seg_0.assign(point_0, point_1)
        self.assertEqual(seg_0.origin().x(), 0)
        self.assertEqual(seg_0.origin().y(), 1)
        self.assertEqual(seg_0.terminal().x(), 10)
        self.assertEqual(seg_0.terminal().y(), 2)
        seg_0 = Segment2D()
        seg_0.assign(point_0, point_1)
        self.assertEqual(seg_0.origin().x(), 0)
        self.assertEqual(seg_0.origin().y(), 1)
        self.assertEqual(seg_0.terminal().x(), 10)
        self.assertEqual(seg_0.terminal().y(), 2)
        seg_0 = Segment2D(point_0, 10, 0)
        seg_0.assign(point_0, point_1)
        self.assertEqual(seg_0.origin().x(), 0)
        self.assertEqual(seg_0.origin().y(), 1)
        self.assertEqual(seg_0.terminal().x(), 10)
        self.assertEqual(seg_0.terminal().y(), 2)

    def test_is_valid(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, 1)
        seg_0 = Segment2D(point_0, point_1)
        self.assertFalse(seg_0.is_valid())

    def test_line(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(10, 10)
        seg_0 = Segment2D(point_0, point_1)
        line = seg_0.line()
        self.assertEqual(abs(line.a() + line.b()) ,0)

    def test_length(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.length(), 9)

    def test_direction(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.direction().degree(), 90)
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, -10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.direction().degree(), -90)

    def test_swap(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1).swap()
        self.assertEqual(seg_0.direction().degree(), -90)
        self.assertEqual(seg_0.origin().y(), 10)

    def test_reverse(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1).reverse()
        self.assertEqual(seg_0.direction().degree(), -90)
        self.assertEqual(seg_0.origin().y(), 10)

    def test_reversed_segment(self):
        point_0 = Vector2D(0, 1)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = seg_0.reversed_segment()
        self.assertEqual(seg_1.direction().degree(), -90)
        self.assertEqual(seg_1.origin().y(), 10)
        seg_0.assign(Vector2D(2, 2), Vector2D(5, 5))
        self.assertEqual(seg_1.direction().degree(), -90)
        self.assertEqual(seg_1.origin().y(), 10)

    def test_perpendicular_bisector(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        p_line = seg_0.perpendicular_bisector()
        self.assertEqual(p_line.get_y(0), 5)

    def test_contains(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.contains(Vector2D(0, 5)))
        self.assertFalse(seg_0.contains(Vector2D(-1, 5)))
        self.assertFalse(seg_0.contains(Vector2D(0, 11)))

    def test_equals(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = Segment2D(point_0, point_1)
        seg_2 = Segment2D(Vector2D(0, 1), point_1)
        self.assertTrue(seg_0 == seg_1)
        self.assertFalse(seg_0 == seg_2)

    def test_equals_weakly(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = Segment2D(Vector2D(0.000000001, 0), point_1)
        self.assertTrue(seg_0.equals_weakly(seg_1))

    def test_projection(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point = seg_0.projection(Vector2D(1, 5))
        self.assertEqual(point.x(), 0)
        self.assertEqual(point.y(), 5)
        point = seg_0.projection(Vector2D(0, 11))
        self.assertFalse(point.is_valid())
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 0)
        seg_0 = Segment2D(point_0, point_1)
        point = seg_0.projection(Vector2D(1, 5))
        self.assertEqual(point.x(), 0)
        self.assertEqual(point.y(), 0)

    def test_intersection(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        seg_1 = Segment2D(point_0, point_1)
        point_i = seg_0.intersection(seg_1)
        self.assertEqual(point_i.x(), 0)
        self.assertEqual(point_i.y(), 1)
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = Segment2D(point_1, point_0)
        point_i = seg_0.intersection(seg_1.line())
        self.assertEqual(point_i.x(), 0)
        self.assertEqual(point_i.y(), 0)
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 12)
        point_1 = Vector2D(-10, 12)
        seg_1 = Segment2D(point_0, point_1)
        point_i = seg_0.intersection(seg_1)
        self.assertEqual(point_i.is_valid(), False)
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = Segment2D(point_0, point_1)
        point_i = seg_0.intersection(seg_1.line())
        self.assertEqual(point_i.is_valid(), False)

    def test_exist_intersection(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.exist_intersection(seg_1))
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.exist_intersection(seg_1))
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 12)
        point_1 = Vector2D(-10, 12)
        seg_1 = Segment2D(point_0, point_1)
        self.assertFalse(seg_0.exist_intersection(seg_1))
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.exist_intersection(seg_1))

    def test_check_intersects_on_line(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.check_intersects_on_line(point_0))
        self.assertTrue(seg_0.check_intersects_on_line(Vector2D(4,5)))
        self.assertTrue(seg_1.check_intersects_on_line(point_1))
        self.assertFalse(seg_1.check_intersects_on_line(Vector2D(20,13)))

    def test_exist_intersection_except_endpoint(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.exist_intersection_except_endpoint(seg_1))
        point_0 = Vector2D(10, 10)
        point_1 = Vector2D(-10, 10)
        seg_1 = Segment2D(point_0, point_1)
        self.assertFalse(seg_0.exist_intersection_except_endpoint(seg_1))

    def test_intersects_except_endpoint(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.intersects_except_endpoint(seg_1))
        point_0 = Vector2D(10, 10)
        point_1 = Vector2D(-10, 10)
        seg_1 = Segment2D(point_0, point_1)
        self.assertFalse(seg_0.intersects_except_endpoint(seg_1))

    def test_nearest_point(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point = seg_0.nearest_point(Vector2D(5, 12))
        self.assertEqual(point.x(), 0)
        self.assertEqual(point.y(), 10)
        point = seg_0.nearest_point(Vector2D(-5, -5))
        self.assertEqual(point.x(), 0)
        self.assertEqual(point.y(), 0)
        point = seg_0.nearest_point(Vector2D(3, 3))
        self.assertEqual(point.x(), 0)
        self.assertEqual(point.y(), 3)

    def test_dist(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.dist(seg_1), 0)
        point_0 = Vector2D(10, -1)
        point_1 = Vector2D(-10, -1)
        seg_1 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.dist(seg_1), 1)
        point_0 = Vector2D(1, 0)
        point_1 = Vector2D(1, 10)
        seg_1 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.dist(seg_1), 1)
        self.assertEqual(seg_0.dist(Vector2D(0, 5)), 0)
        self.assertEqual(seg_0.dist(Vector2D(0, 11)), 1)
        self.assertEqual(seg_0.dist(Vector2D(1, 5)), 1)

    def test_farthest_dist(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertEqual(seg_0.farthest_dist(Vector2D(0, -1)), 11)

    def test_on_segment(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.on_segment(point_0))
        self.assertTrue(seg_0.on_segment(Vector2D(0,5)))
        self.assertFalse(seg_0.on_segment(Vector2D(1,5)))
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        self.assertFalse(seg_0.on_segment(point_0))
        self.assertFalse(seg_0.on_segment(point_1))

    def test_on_segment_weakly(self):
        point_0 = Vector2D(0, 0)
        point_1 = Vector2D(0, 10)
        seg_0 = Segment2D(point_0, point_1)
        self.assertTrue(seg_0.on_segment_weakly(point_0))
        self.assertTrue(seg_0.on_segment_weakly(Vector2D(0.0000001,5.0000001)))
        self.assertFalse(seg_0.on_segment_weakly(Vector2D(0.000001,5.000001)))
        point_0 = Vector2D(10, 1)
        point_1 = Vector2D(-10, 1)
        self.assertFalse(seg_0.on_segment_weakly(point_0))
        self.assertFalse(seg_0.on_segment_weakly(point_1))



""" triangle_2d.py file
    Triangle2D: class name
    Class attributes : _a,_b,_c
    TODO: add test and reverse fix intersection
    TODO: fix the circular import tri-segmant
"""

from __future__ import annotations
# from typing import Union
import math

# from pyrusgeom.segment_2d import Segment2D (removed most likely due to a circular import)
from pyrusgeom.region_2d import Region2D
from pyrusgeom.ray_2d import Ray2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.line_2d import Line2D
from pyrusgeom.math_values import EPSILON


class Triangle2D(Region2D):
    """handling tiiangles in SS2D

    Args:
        Region2D ([type]): created from Region2D class

    Attributes:
        _a: 1st vertex point
        _b: 2nd vertex point
        _c: 3rd vertex point
    """

    def __init__(self, *args) -> None:
        """This is the class init function for Triangle2D

        Args:
            two:
                1 segment 1 vector2D
                (Segment2D, Vector2D)
                Segment2D: segment consist of triangle
                Vector2D: 3rd vertex point
            three:
                (float, float, float)
                float: first vertex point
                float: second vertex point
                float: third vertex point

        Raises:
            Exception: The input must be (three Vector2D) or
                       (one vector 2D and one segment)
        """

        super().__init__()
        if len(args) == 3:
            self._a: Vector2D = Vector2D(args[0])
            self._b: Vector2D = Vector2D(args[1])
            self._c: Vector2D = Vector2D(args[2])
        elif len(args) == 2:  # and isinstance(args[0], Segment2D):
            self._a = Vector2D(args[0].origin_())
            self._b = Vector2D(args[0].terminal_())
            self._c = Vector2D(args[1])
        else:
            raise Exception("The input must be (three Vector2D) or\
                            (one vector 2D and one segment)")

    def assign(self, *args) -> Triangle2D:
        """assign vertex points

        Args:
            two:
                1 segment 1 vector2D
                (Segment2D, Vector2D)
                Segment2D: segment consist of triangle
                Vector2D: 3rd vertex point
            three:
                (float, float, float)
                float: first vertex point
                float: second vertex point
                float: third vertex point

        Raises:
            Exception: The input must be (three Vector2D) or
                       (one vector 2D and one segment)

        Returns:
            Triangle2D: reference to itself
        """
        if len(args) == 3:
            self._a: Vector2D = Vector2D(args[0])
            self._b: Vector2D = Vector2D(args[1])
            self._c: Vector2D = Vector2D(args[2])
        elif len(args) == 2:  # and isinstance(args[0], Segment2D):
            self._a = Vector2D(args[0].origin_())
            self._b = Vector2D(args[0].terminal_())
            self._c = Vector2D(args[1])
        else:
            raise Exception("The input must be (three Vector2D) or\
                            (one vector 2D and one segment)")
        return self

    def is_valid(self) -> bool:
        """check if self triangle is valid or not.

        Returns:
            bool: True if triangle is valid. else False.
        """
        return (self._a.is_valid() and self._b.is_valid() and self._c.is_valid() and
                self._a != self._b and self._b != self._c and self._a != self._a)

    def a(self) -> Vector2D:
        """get 1st point copy

        Returns:
            Vector2D: a Vector2D with point values
        """
        return Vector2D(self._a)

    def b(self) -> Vector2D:
        """get 2nd point copy

        Returns:
            Vector2D: a Vector2D with point values
        """
        return Vector2D(self._b)

    def c(self) -> Vector2D:
        """get 3rd point copy

        Returns:
            Vector2D: a Vector2D with point values
        """
        return Vector2D(self._c)

    def a_(self) -> Vector2D:
        """get 1st point

        Returns:
            Vector2D: reference to the member variable
        """
        return self._a

    def b_(self) -> Vector2D:
        """get 2nd point

        Returns:
            Vector2D: reference to the member variable
        """
        return self._b

    def c_(self) -> Vector2D:
        """get 3rd point

        Returns:
            Vector2D: reference to the member variable
        """
        return self._c

    def area(self) -> float:
        """get the area of self region

        Returns:
            float: value of the area
        """
        return math.fabs((self._b - self._a).outer_product(self._c - self._a)) * 0.5

    def signed_area(self) -> float:
        """get a signed area.

        method is equivalent to signed_area_st().

        If points a, b, are placed counterclockwise order, positive number.
        If points a, b, are placed clockwise order, negative number.
        If points a, b, are placed on a line, 0.

        Returns:
            float: signed area value
        """
        return Triangle2D.signed_area_st(self._a, self._b, self._c)

    def double_signed_area(self) -> float:
        """get a double of signed area value.

        method is equivalent to double_signed_area_st().

        If points a, b, are placed counterclockwise order, positive number.
        If points a, b, are placed clockwise order, negative number.
        If points a, b, are placed on a line, 0.

        Returns:
            float: double of signed area value
        """
        return Triangle2D.double_signed_area_st(self._a, self._b, self._c)

    def ccw(self) -> bool:
        """check if self triangle's vertices are placed counterclockwise order.

        Returns:
            bool: True if counterclockwise. else False
        """
        return Triangle2D.tri_ccw(self._a, self._b, self._c)

    def contains(self, point: Vector2D) -> bool:
        """check if self triangle contains 'point'.

        Args:
            point (Vector2D): considered point

        Returns:
            bool: True if contains. else False.
        """
        rel1 = Vector2D(self._a - point)
        rel2 = Vector2D(self._b - point)
        rel3 = Vector2D(self._c - point)

        outer1 = rel1.outer_product(rel2)
        outer2 = rel2.outer_product(rel3)
        outer3 = rel3.outer_product(rel1)

        if (outer1 >= 0.0 and outer2 >= 0.0 and
            outer3 >= 0.0) or (outer1 <= 0.0 and
                               outer2 <= 0.0 and outer3 <= 0.0):
            return True
        return False

    def centroid(self) -> Vector2D:
        """get the center of gravity(centroid)

        Returns:
            Vector2D: coordinates of gravity center
        """
        return Triangle2D.tri_centroid(self._a, self._b, self._c)

    def incenter(self) -> Vector2D:
        """get the center of inscribed circle

        Returns:
            Vector2D: coordinates of inner center
        """
        return Triangle2D.tri_incenter(self._a, self._b, self._c)

    def circumcenter(self) -> Vector2D:
        """get the center of circumscribed circle

        Returns:
            Vector2D: coordinates of outer center
        """
        return Triangle2D.tri_circumcenter(self._a, self._b, self._c)

    def orthocenter(self) -> Vector2D:
        """get the orthocenter

        Returns:
            Vector2D: coordinates of orthocenter
        """
        return Triangle2D.tri_orthocenter(self._a, self._b, self._c)

    def intersection(self, other): # Union[Line2D, Ray2D, Segment2D]) -> list:
        """
        TODO: Need to rewrite. fixed TriBug
            Line2D
        brief calculate intersection point with line.
        param line considered line.
        return number of intersection + sol 1 + sol 2
            Ray2D
        brief calculate intersection point with ray.
        param ray considered ray line.
        return number of intersection + sol 1 + sol 2
            Segment2D
        brief calculate intersection point with line segment.
        param segment considered line segment.
        return number of intersection + sol 1 + sol 2
        """
        if isinstance(other, Line2D):
            n_sol = 0
            t_sol = [Vector2D(), Vector2D()]
            from pyrusgeom.segment_2d import Segment2D
            t_sol[n_sol] = Segment2D(self._a, self._b).intersection(other)
            if n_sol < 2 and t_sol[n_sol].is_valid():
                n_sol += 1

            t_sol[n_sol] = Segment2D(self._b, self._c).intersection(other)
            if n_sol < 2 and t_sol[n_sol].is_valid():
                n_sol += 1

            t_sol[n_sol] = Segment2D(self._c, self._a).intersection(other)
            if n_sol < 2 and t_sol[n_sol].is_valid():
                n_sol += 1

            if n_sol == 2 and math.fabs(t_sol[0].x() - t_sol[1].x()) < EPSILON and math.fabs(
                    t_sol[0].y() - t_sol[1].y()) < EPSILON:
                n_sol = 1
            sol_list = [n_sol, t_sol[0], t_sol[1]]

            return sol_list
        if isinstance(other, Ray2D):
            n_sol = self.intersection(other.line())

            if n_sol[0] > 1 and not other.in_right_dir(n_sol[1], 1.0):
                n_sol[0] -= 1

            if n_sol[0] > 0 and not other.in_right_dir(n_sol[1], 1.0):
                n_sol[1] = n_sol[2]
                n_sol[0] -= 1

            return n_sol

        if isinstance(other, Segment2D):
            n_sol = self.intersection(other.line())

            if n_sol[0] > 1 and not other.contains(n_sol[2]):
                n_sol[0] -= 1

            if n_sol > 0 and not other.contains(n_sol[1]):
                n_sol[1] = n_sol[2]
                n_sol[0] -= 1

            return n_sol
        return [0]

    # static methods

    @staticmethod
    def double_signed_area_st(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> float:
        """get a double signed area value (area of parallelogram)

        If points a, b, are placed counterclockwise order, positive number.
        If points a, b, are placed clockwise order, negative number.
        If points a, b, are placed on a line, 0.

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            float: double singed area value.
        """
        return ((v_a.x() - v_c.x()) * (v_b.y() - v_c.y())
                + (v_b.x() - v_c.x()) * (v_c.y() - v_a.y()))

    @staticmethod
    def signed_area_st(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> float:
        """get a signed area value

        If points a, b, are placed counterclockwise order, positive number.
        If points a, b, are placed clockwise order, negative number.
        If points a, b, are placed on a line, 0.

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            float: signed area value
        """
        return Triangle2D.double_signed_area_st(v_a, v_b, v_c) * 0.5

    @staticmethod
    def tri_ccw(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> bool:
        """check if input vertices are placed counterclockwise order.

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            bool: True if counterclockwise. else False
        """
        return Triangle2D.double_signed_area_st(v_a, v_b, v_c) > 0.0

    @staticmethod
    def tri_centroid(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> Vector2D:
        """ get the center of gravity

        centroid = (a + b + c) / 3

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            Vector2D: coordinates of gravity center
        """
        return Vector2D(v_a).add(v_b).add(v_c) / 3.0

    @staticmethod
    def tri_incenter(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> Vector2D:
        """get the incenter point

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            Vector2D: coordinates of incenter
        """
        a_b = v_b - v_a
        a_c = v_c - v_a
        bisect_a = Line2D(v_a, AngleDeg.bisect(a_b.th(), a_c.th()))

        b_a = v_a - v_b
        b_c = v_c - v_b
        bisect_b = Line2D(v_b, AngleDeg.bisect(b_a.th(), b_c.th()))

        return bisect_a.intersection(bisect_b)

    @staticmethod
    def tri_circumcenter(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> Vector2D:
        """get the circumcenter point

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            Vector2D: coordinates of circumcenter
        """
        perpendicular_ab = Line2D.perpendicular_bisector(v_a, v_b)
        perpendicular_bc = Line2D.perpendicular_bisector(v_b, v_c)

        sol = perpendicular_ab.intersection(perpendicular_bc)

        if not sol.is_valid():
            perpendicular_ca = Line2D.perpendicular_bisector(v_c, v_a)
            sol = perpendicular_ab.intersection(perpendicular_ca)

            if sol.is_valid():
                return sol

            sol = perpendicular_bc.intersection(perpendicular_ca)
            if sol.is_valid():
                return sol

        a_b = v_b - v_a
        c_a = v_c - v_a

        tmp = a_b.outer_product(c_a)
        if math.fabs(tmp) < 1.0e-10:  # The area of parallelogram is 0.
            return Vector2D.invalid()

        inv = 0.5 / tmp
        ab_len2 = a_b.r2()
        ca_len2 = c_a.r2()
        xcc = inv * (ab_len2 * c_a.get_y() - ca_len2 * a_b.get_y())
        ycc = inv * (a_b.get_x() * ca_len2 - c_a.get_x() * ab_len2)

        return Vector2D(v_a.x() + xcc, v_a.y() + ycc)

    @staticmethod
    def tri_orthocenter(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D) -> Vector2D:
        """get the orthocenter point

        orthocenter = a + b + c - 2 * circumcenter

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point

        Returns:
            Vector2D: coordinates of orthocenter
        """
        perpend_a = Line2D(v_b, v_c).perpendicular(v_a)
        perpend_b = Line2D(v_c, v_a).perpendicular(v_b)
        return perpend_a.intersection(perpend_b)

    @staticmethod
    def tri_contains(v_a: Vector2D, v_b: Vector2D, v_c: Vector2D, point: Vector2D) -> bool:
        """ check if triangle(a,b,c) contains the input point.

        Args:
            v_a (Vector2D): triangle's 1st vertex point
            v_b (Vector2D): triangle's 2nd vertex point
            v_c (Vector2D): triangle's 3rd vertex point
            point (Vector2D): checked point

        Returns:
            bool: True if contains. else Fasle.
        """
        rel1 = Vector2D(v_a - point)
        rel2 = Vector2D(v_b - point)
        rel3 = Vector2D(v_c - point)

        outer1 = rel1.outer_product(rel2)
        outer2 = rel2.outer_product(rel3)
        outer3 = rel3.outer_product(rel1)

        if (outer1 >= 0.0 and outer2 >= 0.0 and
            outer3 >= 0.0) or (outer1 <= 0.0 and
                               outer2 <= 0.0 and outer3 <= 0.0):
            return True
        return False

    def __repr__(self) -> str:
        """represent Triangle2D as a string

        Returns:
            str: Triangle2D's _a, _b and _c as string
        """
        return f"[{self._a},{self._b},{self._c}]"

    def to_str(self, ostr: str) -> str:
        """add triangle to the ostr

        Args:
            ostr (str): str to add to it
        """
        ostr += f'(tri {round(self.a().x(), 3)} {round(self.a().y(), 3)} {round(self.b().x(), 3)} \
        {round(self.b().y(), 3)} {round(self.c().x(), 3)} {round(self.c().y(), 3)})'



from unittest import TestCase
from pyrusgeom.line_2d import Line2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg


class TestLine2D(TestCase):
    def test_general(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        self.assertEqual(line.a(), -10)
        self.assertEqual(line.b(), 0)
        self.assertEqual(line.c(), 0)

        line = Line2D(Vector2D(0, 0), 90)
        self.assertEqual(line.a(), -1)
        self.assertTrue(abs(line.b()) < 0.01)
        self.assertEqual(line.c(), 0)

    def test_get_x(self):
        line = Line2D(Vector2D(0, 0), Vector2D(10, 10))
        self.assertEqual(line.get_x(5), 5)

    def test_get_y(self):
        line = Line2D(Vector2D(0, 0), Vector2D(10, 10))
        self.assertEqual(line.get_y(5), 5)

    def test_copy(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        l2 = line.copy()
        self.assertEqual(l2.a(), -10)
        self.assertEqual(l2.b(), 0)
        self.assertEqual(l2.c(), 0)

    def test_dist(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        self.assertEqual(line.dist(Vector2D(0, -1)), 0)
        self.assertEqual(line.dist(Vector2D(2, 2)), 2)

    def test_dist2(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        self.assertEqual(line.dist2(Vector2D(0, -1)), 0)
        self.assertEqual(line.dist2(Vector2D(2, 2)), 4)

    def test_is_parallel(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        l2 = Line2D(Vector2D(1, 0), Vector2D(1, -10))
        self.assertTrue(line.is_parallel(l2))

    def test_intersection(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        l2 = Line2D(Vector2D(1, 0), Vector2D(1, -10))
        self.assertFalse(line.intersection(l2).is_valid())
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        l2 = Line2D(Vector2D(1, 1), Vector2D(-1, 1))
        self.assertTrue(line.intersection(l2).x() == 0)
        self.assertTrue(line.intersection(l2).y() == 1)

    def test_perpendicular(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        p = Vector2D(5, 5)
        per = line.perpendicular(p)
        self.assertEqual(per.get_y(2), 5)

    def test_projection(self):
        line = Line2D(Vector2D(0, 0), Vector2D(0, 10))
        p = Vector2D(5, 5)
        pro = line.projection(p)
        self.assertEqual(pro.x(), 0)
        self.assertEqual(pro.y(), 5)

    def test_line_intersection(self):
        pass

    def test_angle_bisector(self):
        line = Line2D.angle_bisector(Vector2D(1, 1), 0, 90)
        self.assertTrue(abs(line.get_x(5) - 5) < 0.01)

    def test_perpendicular_bisector(self):
        line = Line2D.perpendicular_bisector(Vector2D(-5, 0), Vector2D(5, 0))
        self.assertEqual(line.get_x(5), 0)

    def test_kwargs(self):
        line = Line2D(p1=Vector2D(0, 0), p2=Vector2D(0, 10))
        self.assertEqual(line.a(), -10)
        self.assertEqual(line.b(), 0)
        self.assertEqual(line.c(), 0)

        line = Line2D(p=Vector2D(0, 0), a=90)
        self.assertEqual(line.a(), -1)
        self.assertTrue(abs(line.b()) < 0.01)
        self.assertEqual(line.c(), 0)

        line2 = Line2D(line)

        self.assertEqual(line2.a(), -1)
        self.assertTrue(abs(line2.b()) < 0.01)
        self.assertEqual(line2.c(), 0)

        line = Line2D(p=Vector2D(0, 0), a=AngleDeg(90))
        self.assertEqual(line.a(), -1)
        self.assertTrue(abs(line.b()) < 0.01)
        self.assertEqual(line.c(), 0)



""" line_2d.py file
    Line2D: class name
    Line Formula: aX + bY + c = 0
"""
from __future__ import annotations
import math

from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.math_values import EPSILON, ERROR_VALUE


class Line2D:
    """ handling 2D straight line is SS2D.
    Attributes:
        _a,_b,_c
        Line Formula: aX + bY + c = 0
    """

    def __init__(self, *args, **kwargs) -> None:
        """This is the class init function for Line2D

        Args:
            one:
                copy from line
                Line2D: line
            two:
                constructed from 2 points
                (Vector2D, Vector2D)
                Vector2D: first point
                Vector2D: second point

                constructed from origin point + direction
                (Vector2D, AngleDeg)
                Vector2D: org origin point
                AngleDeg: line dir direction from origin point
            three:
                (float, float, float)
                float: assigned a value
                float: assigned b value
                float: assigned c value

        Raises:
            Exception: The input must be (two Vector2D) or (3 float) or
                        (one vector 2D and one angle)
        """
        self._a = 0.0
        self._b = 0.0
        self._c = 0.0
        self._is_valid = False
        if len (kwargs) == 0:
            if len(args) == 3:
                if isinstance(args[0], (float, int)) and \
                        isinstance(args[1], (float, int)) and \
                        isinstance(args[2], (float, int)):
                    self._a = args[0]
                    self._b = args[1]
                    self._c = args[2]
                    self._is_valid = True
            elif len(args) == 2:
                if isinstance(args[0], Vector2D) and isinstance(args[1], (AngleDeg, float, int)):
                    line_dir = args[1] if isinstance(
                        args[1], AngleDeg) else AngleDeg(args[1])
                    self._a = -line_dir.sin()
                    self._b = line_dir.cos()
                    self._c = -self._a * args[0].x() - self._b * args[0].y()
                    self._is_valid = True
                elif isinstance(args[0], Vector2D) and isinstance(args[1], Vector2D):
                    self._a = -(args[1].y() - args[0].y())
                    self._b = args[1].x() - args[0].x()
                    self._c = -self._a * args[0].x() - self._b * args[0].y()
                    self._is_valid = True
            elif len(args) == 1:
                if isinstance(args[0], Line2D):
                    self._a = args[0].a()
                    self._b = args[0].b()
                    self._c = args[0].c()
                    self._is_valid = True
        else:
            if {'a','b','c'} <= set(kwargs):
                self._a = kwargs['a']
                self._b = kwargs['b']
                self._c = kwargs['c']
                self._is_valid = True
            elif {'p1','p2'} <= set(kwargs):
                self._a = -(kwargs['p2'].y() - kwargs['p1'].y())
                self._b = kwargs['p2'].x() - kwargs['p1'].x()
                self._c = -self._a * kwargs['p1'].x() - self._b * kwargs['p1'].y()
                self._is_valid = True
            elif {'p','a'} <= set(kwargs):
                line_dir = kwargs['a'] if isinstance(
                        kwargs['a'], AngleDeg) else AngleDeg(kwargs['a'])
                self._a = -line_dir.sin()
                self._b = line_dir.cos()
                self._c = -self._a * kwargs['p'].x() - self._b * kwargs['p'].y()
                self._is_valid = True
            elif 'l' in kwargs:
                self._a = kwargs['l'].a()
                self._b = kwargs['l'].b()
                self._c = kwargs['l'].c()
                self._is_valid = True
        if not self._is_valid:
            raise Exception('The input should be(2 vector2d) or (3 float) \
                            or (vector 2d and angle)')

    def a(self) -> float:
        """accessor coefficient 'A'

        Returns:
            float:  coefficient 'A'  of line formula
        """
        return self._a

    def b(self) -> float:
        """accessor coefficient 'B'

        Returns:
            float:  coefficient 'B'  of line formula
        """
        return self._b

    def c(self) -> float:
        """accessor coefficient 'C'

        Returns:
            float:  coefficient 'C'  of line formula
        """
        return self._c

    def get_x(self, f_y: float) -> float:
        """get X-coordinate correspond to 'f_y'

        Args:
            f_y (float): considered Y

        Returns:
            float: X coordinate
        """
        if math.fabs(self._a) < EPSILON:
            return ERROR_VALUE
        return -(self._b * f_y + self._c) / self._a

    def get_y(self, f_x) -> float:
        """get Y-coordinate correspond to 'f_x'

        Args:
            f_x (float): considered X

        Returns:
            float: Y coordinate
        """
        if math.fabs(self._b) < EPSILON:
            return ERROR_VALUE
        return -(self._a * f_x + self._c) / self._b

    def copy(self) -> Line2D:
        """return a copy of this line

        Returns:
            Line2D: copy of this line
        """
        return Line2D(self._a, self._b, self._c)

    def dist(self, point: Vector2D) -> float:
        """calculate distance from point to this line

        Args:
            point (Vector2D): considered point

        Returns:
            float: distance value
        """
        return math.fabs(
            (self._a * point.x() + self._b * point.y() + self._c) /
            math.sqrt(self._a * self._a + self._b * self._b))

    def dist2(self, point: Vector2D) -> float:
        """get squared distance from this line to point

        Args:
            point (Vector2D): considered point

        Returns:
            float: squared distance value
        """
        delta = self._a * point.x() + self._b * point.y() + self._c
        return (delta * delta) / (self._a * self._a + self._b * self._b)

    def is_parallel(self, other: Line2D) -> bool:
        """check if the slope of this line is parallel to the slope of 'other'

        Args:
            other (Line2D): considered line

        Returns:
            bool: true if almost parallel. else false
        """
        return math.fabs(self._a * other.b() - other.a() * self._b) < EPSILON

    def intersection(self, other: Line2D) -> Vector2D:
        """get the intersection point with 'other' line

        Args:
            other (Line2D): considered line

        Returns:
            Vector2D: intersection point. if it does not exist,
             the invalidated value vector is returned.
        """
        return Line2D.line_intersection(self, other)

    def perpendicular(self, point: Vector2D) -> Line2D:
        """calculate perpendicular line

        Args:
            point (Vector2D): the point that perpendicular line pass through

        Returns:
            Line2D: perpendicular line
        """
        return Line2D(self._b, -self._a, self._a * point.y() - self._b * point.x())

    def projection(self, point: Vector2D) -> Vector2D:
        """calculate projection point from point

        Args:
            point (Vector2D): base point

        Returns:
            Vector2D: projection point
        """
        return self.intersection(self.perpendicular(point))

    @staticmethod
    def line_intersection(line1: Line2D, line2: Line2D) -> Vector2D:
        """get the intersection point of 2 lines

        Args:
            line1 (Line2D): the first line
            line2 (Line2D): the second line

        Returns:
            Vector2D: the intersection point. if no intersection, invalidated vector is returned.
        """
        tmp = line1.a() * line2.b() - line1.b() * line2.a()
        if math.fabs(tmp) < EPSILON:
            return Vector2D.invalid()

        return Vector2D((line1.b() * line2.c() - line2.b() * line1.c()) / tmp,
                        (line2.a() * line1.c() - line1.a() * line2.c()) / tmp)

    @staticmethod
    def angle_bisector(origin: Vector2D, left: AngleDeg, right: AngleDeg) -> Line2D:
        """make angle bisector line from two angles

        Args:
            origin (Vector2D): point that is passed through by result line
            left (AngleDeg): left angle
            right (AngleDeg):  right angle

        Returns:
            Line2D: line object
        """
        return Line2D(origin, AngleDeg.bisect(left, right))

    @staticmethod
    def perpendicular_bisector(point1: Vector2D, point2: Vector2D) -> Line2D:
        """make perpendicular bisector line from twt points

        Args:
            point1 (Vector2D): 1st point
            point2 (Vector2D): 2nd point

        Returns:
            Line2D: line object
        """
        if (math.fabs(point2.x() - point1.x()) < EPSILON and
                math.fabs(point2.y() - point1.y()) < EPSILON):
            print("Error : points have same coordinate values")
            tmp_vec = Vector2D(point1.x() + 1, point2.y())
            return Line2D(point1, tmp_vec)
        tmp = (point2.x() * point2.x() - point1.x() * point1.x()
               + point2.y() * point2.y() - point1.y() * point1.y()) * -0.5
        return Line2D(point2.x() - point1.x(), point2.y() - point1.y(), tmp)

    def __eq__(self, other: Line2D) -> bool:
        """operator == for Line2D

        Args:
            other (Line2D): right hand side argument
        Returns:
            bool: true if equal or difference is less than EPSILON. else false
        """
        return (math.fabs(self._a * other.b() - other.a() * self._b)
                < EPSILON and self._c == other.c())

    def __repr__(self) -> str:
        """represent Line2D as a string
           aX + bY + c = 0
        Returns:
            str: Line as string
        """

        if self._c == 0:
            return f"({self._a} X + {self._b} Y = 0)"
        return f"({self._a} X + {self._b} Y + {self._c} = 0)"



""" file matrix_2d.py
    Matrix2D: class name
"""

from __future__ import annotations
from typing import Union
import math
from pyrusgeom.math_values import EPSILON

from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg


class Matrix2D:
    """ 2D transform matrix class File.

        ( m11, m12, dxtf )
        ( m21, m22, dytf )
        (   0,   0,  1 )

    Attributes:
        self.11 element (1,1): the horizontal scaling factor.
        self.12 element (1,2): the vertical shearing factor.
        self.21 element (2,1): the horizontal shearing factor.
        self.22 element (2,2): the vertical scaling factor.
        self.dxtf dxtf: the horizontal translation factor.
        self.dytf dytf: the vertical translation factor.

    """
    _m11:float
    _m12:float
    _dx:float
    _m21:float
    _m22:float
    _dy:float


    def __init__(self, __11=1.0, __12=0.0, __21=0.0, __22=1.0, __x=0.0, __y=0.0):
        """This is the class init function and creates a matrix2d

        Defualt:
            create a matrix with all the given elements
        OR
            create an identity matrix
        Args:
            __11 (float, optional): the horizontal scaling factor. Defaults to 1.0.
            __12 (float, optional): the vertical shearing factor. Defaults to 0.0.
            __21 (float, optional): the horizontal shearing factor. Defaults to 0.0.
            __22 (float, optional): the vertical scaling factor. Defaults to 1.0.
            __x (float, optional): the horizontal translation factor. Defaults to 0.0.
            __y (float, optional): the vertical translation factor. Defaults to 0.0.
        """
        self._m11 = __11
        self._m12 = __12
        self._m21 = __21
        self._m22 = __22
        self._dx = __x
        self._dy = __y
        self.is_valid = True

    def reset(self):
        """ reset this matrix to the identity matrix
        """
        self._m11 = self._m22 = 1.0
        self._m12 = self._m21 = self._dx = self._dy = 0.0

    def assign(self, __11: float, __12: float, __21: float, __22: float, __x: float, __y: float):
        """set this matrix elements with the given specified values.

        Args:
            __11 (float): the horizontal scaling factor.
            __12 (float): the vertical shearing factor.
            __21 (float): the horizontal shearing factor.
            __22 (float): the vertical scaling factor.
            __x (float): the horizontal translation factor.
            __y (float): the vertical translation factor.
        """
        self._m11 = __11
        self._m12 = __12
        self._m21 = __21
        self._m22 = __22
        self._dx = __x
        self._dy = __y

    def m11(self) -> float:
        """get the horizontal scaling factor.

        Returns:
            float: the horizontal scaling factor value.
        """
        return self._m11

    def m12(self) -> float:
        """get the vertical shearing factor.

        Returns:
            float: return the vertical shearing factor value.
        """
        return self._m12

    def m21(self) -> float:
        """get the horizontal shearing factor.

        Returns:
            float: the horizontal shearing factor value.
        """
        return self._m21

    def m22(self) -> float:
        """get the vertical scaling factor.

        Returns:
            float: the vertical scaling factor value.
        """
        return self._m22

    def dxtf(self) -> float:
        """get the horizontal translation factor.

        Returns:
            float: the horizontal translation factor value.
        """
        return self._dx

    def dytf(self) -> float:
        """get the vertical translation factor.

        Returns:
            float: the vertical translation factor value.
        """
        return self._dy

    # def __members(self):
    #     """get the matrix members

    #     Returns:
    #         object: (m11,m12,dx,m21,m22,dy)
    #     """
    #     return (self.m11(),self.m12(),self.dxtf(),self.m21(),self.m22(),self.dytf)

    def det(self) -> float:
        """get the matrix's determinant

        Returns:
            float: the determinant value.
        """
        return self._m11 * self._m22 - self._m12 * self._m21

    def invertible(self) -> bool:
        """check if this matrix is invertible (is not insular).

        Returns:
            bool: true if this matrix is invertible.else false.
        """
        return math.fabs(self.det()) > EPSILON

    def inverted(self) -> Matrix2D:
        """get the inverted matrix.

        if the matrix isn't invertible return the defualt Matrix

        Returns:
            Matrix2D: if invertible return the inverted matrix object.
        """
        determinant = self.det()
        if determinant == 0.0:  # not invertible
            return Matrix2D()  # default matrix

        dinv = 1.0 / determinant
        return Matrix2D(self._m22 * dinv, -self._m12 * dinv,
                        -self._m21 * dinv, self._m11 * dinv,
                        (self._m12 * self._dy - self._dx * self._m22) * dinv,
                        (self._dx * self._m21 - self._m11 * self._dy) * dinv)

    def translate(self, dxtf: float, dytf: float):
        """moves the coordinate as the other matrix.

        SameAs:
            self = Matrix2D.make_translation(dxtf,dytf) * self

        Args:
            dxtf (float): move factor for the x axis.
            dytf (float): move factor for the y axis.
        """
        self._dx += dxtf
        self._dy += dytf

    def scale(self, scale_x: float, scale_y: float):
        """scales the coordinate.

        SameAs:
            self = Matrix2D.make_scaling(scale_x,scale_y) * self

        Args:
            scale_x (float): scaling factor for the x axis.
            scale_y (float): scaling factor for the y axis.
        """
        self._m11 *= scale_x
        self._m12 *= scale_x
        self._dx *= scale_x
        self._m21 *= scale_y
        self._m22 *= scale_y
        self._dy *= scale_y

    def rotate(self, angle: AngleDeg):
        """rotates the coordinate system

        SameAs:
            self = Matrix2D.make_rotation(angle) * self

        Args:
            angle (AngleDeg): rotation angle
        """
        ang_sin = angle.sin()
        ang_cos = angle.cos()
        tm11 = self._m11 * ang_cos - self._m21 * ang_sin
        tm12 = self._m12 * ang_cos - self._m22 * ang_sin
        tm21 = self._m11 * ang_sin + self._m21 * ang_cos
        tm22 = self._m12 * ang_sin + self._m22 * ang_cos
        tdx = self._dx * ang_cos - self._dy * ang_sin
        tdy = self._dx * ang_sin + self._dy * ang_cos

        self._m11 = tm11
        self._m12 = tm12
        self._dx = tdx
        self._m21 = tm21
        self._m22 = tm22
        self._dy = tdy

    def transform(self, *args) -> Vector2D:
        """create transformed vector from input

        Args:
            One:
                vector (Vector2D): input vector
            Two:
                x (float): input x-coordinates value
                y (folat): input y-coordinates value
        Raises:
            Exception: Need one Vector2D or Two floats
                to pervent raise

        Returns:
            Vector2D : mapped vector object
        """
        if len(args) == 1:
            vector:Vector2D = args[0]
            return Vector2D(self._m11 * vector.x() + self._m12 * vector.y() + self._dx,
                            self._m21 * vector.x() + self._m22 * vector.y() + self._dy)
        if len(args) == 2:
            return Vector2D(self._m11 * args[0] + self._m12 * args[1] + self._dx,
                            self._m21 * args[0] + self._m22 * args[1] + self._dy)

        raise Exception(
            'The input should must inclue a Vector2D or two float for xy')

    def transform_vec(self, vector: Vector2D) -> Vector2D:
        """transform input vector with this matrix

        Args:
            vector (Vector2D): input vector

        Returns:
            Vector2D: trasformed vector
        """
        t_x = self._m11 * vector.x() + self._m12 * vector.y() + self._dx
        t_y = self._m21 * vector.x() + self._m22 * vector.y() + self._dy
        vector.assign(t_x, t_y)
        return vector

    @staticmethod
    def make_translation(dxtf: float, dytf: float) -> Matrix2D:
        """create the translation matrix.

        Args:
            dxtf (float): the horizontal translation factor.
            dytf (float): the vertical translation factor.

        Returns:
            Matrix2D: new matrix object
        """
        return Matrix2D(1.0, 0.0, 0.0, 1.0, dxtf, dytf)

    @staticmethod
    def make_scaling(scale_x: float, scale_y: float) -> Matrix2D:
        """create the scaling matrix.

        Args:
            scale_x (float): the horizontal scaling factor.
            scale_y (float): the vertical scaling factor.

        Returns:
            Matrix2D: new matrix object
        """
        return Matrix2D(scale_x, 0.0, 0.0, scale_y, 0.0, 0.0)

    @staticmethod
    def make_rotation(angle: AngleDeg) -> Matrix2D:
        """create the rotation matrix.

        Args:
            angle (AngleDeg): angle the rotation angle

        Returns:
            Matrix2D: new matrix object
        """
        ang_cos = angle.cos()
        ang_sin = angle.sin()
        return Matrix2D(ang_cos, -ang_sin, ang_sin, ang_cos, 0.0, 0.0)

    def __imul__(self, other: Matrix2D):
        """multiplied by other matrix

        Args:
            other ([Matrix2D]): left hand side matrix
        """
        tm11 = self._m11 * other.m11() + self._m12 * other.m21()
        tm12 = self._m11 * other.m12() + self._m12 * other.m22()
        tm21 = self._m21 * other.m11() + self._m22 * other.m21()
        tm22 = self._m21 * other.m12() + self._m22 * other.m22()
        tdx = self._m11 * other.dxtf() + self._m12 * other.dytf() + self._dx
        tdy = self._m21 * other.dxtf() + self._m22 * other.dytf() + self._dy

        self._m11 = tm11
        self._m12 = tm12
        self._m21 = tm21
        self._m22 = tm22
        self._dx = tdx
        self._dy = tdy

        return self

    def __mul__(self, other: Union[Matrix2D, Vector2D]) -> Union[Matrix2D, Vector2D]:
        """multiplication operator of Matrix * Matrix OR
        multiplication(transformation) operator of Matrix x Vector.

        Args:
            other (Union[Matrix2D, Vector2D]): right hand side matrix OR vector

        Returns:
            Union[Matrix2D, Vector2D]: result matrix object or vector opject
        """
        if isinstance(other, Vector2D):
            return self.transform(other)
        if isinstance(other, Matrix2D):
            mat_tmp = self
            return mat_tmp.__imul__(other)
        return self

    # def __hash__(self):
    #     return hash(self.__members())

    def __eq__(self, other: Matrix2D) -> bool:
        """__eq__ operator

        Args:
            other (Matrix2D): other matrix to compare

        Returns:
            bool: true if eq. else false
        """

        return self._m11 == other.m11() and self._m12 == other.m12() and\
            self._m21 == other.m21() and self._m22 == other.m22() and\
            self._dx == other.dytf() and self._dy == other.dytf()


    def __repr__(self):
        """represent the Matrix2D as a logical string

        Returns:
            str: contains Matrix2D index
        """
        return f"{[self._m11, self._m12, self._m21, self._m22, self._dx, self._dy]}"



""" test_angle_deg.py file
    to test pyrusgeom AngleDeg class
"""
from unittest import TestCase
from pyrusgeom.angle_deg import AngleDeg


class TestAngleDeg(TestCase):
    """TestAnlgeDeg class

    Args:
        TestCase (UnitTest): fail if any of tests falis
    """
    def test_degree(self):
        alpha_angle = AngleDeg()
        self.assertEqual(alpha_angle.degree(), 0)
        alpha_angle = AngleDeg(10)
        self.assertEqual(alpha_angle.degree(), 10)
        alpha_angle = AngleDeg(-90)
        self.assertEqual(alpha_angle.degree(), -90)
        alpha_angle = AngleDeg(360)
        self.assertEqual(alpha_angle.degree(), 0)
        alpha_angle = AngleDeg(-360)
        self.assertEqual(alpha_angle.degree(), 0)
        alpha_angle = AngleDeg(AngleDeg(10))
        self.assertEqual(alpha_angle.degree(), 10)
        alpha_angle = AngleDeg(af=-90)
        self.assertEqual(alpha_angle.degree(), -90)


    def test_set_degree(self):
        alpha_angle = AngleDeg(20)
        alpha_angle.set_degree(30)
        self.assertEqual(alpha_angle.degree(), 30)

    def test_set_angle(self):
        alpha_angle = AngleDeg(20)
        beta_angle = AngleDeg(30)
        alpha_angle.set_angle(beta_angle)
        self.assertEqual(alpha_angle.degree(), 30)
        beta_angle.set_degree(10)
        self.assertEqual(alpha_angle.degree(), 30)

    def test_is_within(self):
        alpha_angle = AngleDeg(30)
        beta_angle = AngleDeg(50)
        gamma_angle = AngleDeg(40)
        self.assertTrue(gamma_angle.is_within(alpha_angle, beta_angle))
        self.assertFalse(gamma_angle.is_within(beta_angle, alpha_angle))

    def test_abs(self):
        alpha_angle = AngleDeg(-20)
        self.assertEqual(alpha_angle.abs(), 20)

    def test_radian(self):
        alpha_angle = AngleDeg(30)
        self.assertAlmostEqual(alpha_angle.radian(), 0.523599 , 6)

    def test_reverse(self):
        alpha_angle = AngleDeg(-20)
        alpha_angle.reverse()
        self.assertEqual(alpha_angle.degree(), 160)
        self.assertEqual(alpha_angle.degree(), 160)

    def test_reverse_angle(self):
        alpha_angle = AngleDeg(-20)
        self.assertEqual(alpha_angle.reverse_angle().degree(), 160)
        self.assertNotEqual(alpha_angle.degree(), 160)

    def test_is_left_of(self):
        alpha_angle = AngleDeg(20)
        beta_angle = AngleDeg(30)
        gamma_angle = AngleDeg(30)
        delta = AngleDeg(45)
        self.assertFalse(gamma_angle.is_left_of(beta_angle))
        self.assertTrue(gamma_angle.is_left_of(delta))
        self.assertFalse(gamma_angle.is_left_of(alpha_angle))

    def test_is_right_of(self):
        alpha_angle = AngleDeg(20)
        beta_angle = AngleDeg(30)
        gamma_angle = AngleDeg(30)
        delta = AngleDeg(45)
        self.assertFalse(gamma_angle.is_right_of(beta_angle))
        self.assertFalse(gamma_angle.is_right_of(delta))
        self.assertTrue(gamma_angle.is_right_of(alpha_angle))

    def test_is_left_equal_of(self):
        alpha_angle = AngleDeg(20)
        beta_angle = AngleDeg(30)
        gamma_angle = AngleDeg(30)
        delta = AngleDeg(45)
        self.assertTrue(gamma_angle.is_left_equal_of(beta_angle))
        self.assertTrue(gamma_angle.is_left_equal_of(delta))
        self.assertFalse(gamma_angle.is_left_equal_of(alpha_angle))

    def test_is_right_equal_of(self):
        alpha_angle = AngleDeg(20)
        beta_angle = AngleDeg(30)
        gamma_angle = AngleDeg(30)
        delta = AngleDeg(45)
        self.assertTrue(gamma_angle.is_right_equal_of(beta_angle))
        self.assertFalse(gamma_angle.is_right_equal_of(delta))
        self.assertTrue(gamma_angle.is_right_equal_of(alpha_angle))
    def test_copy(self):
        alpha_angle = AngleDeg(10)
        beta_angle = alpha_angle.copy()
        beta_angle.set_degree(20)
        self.assertNotEqual(alpha_angle.degree(), 20)

    def test_get_normalized(self):
        alpha_angle = AngleDeg(90)
        self.assertEqual(alpha_angle.get_normalized(), 0.75)

    def test_cos(self):
        alpha_angle = AngleDeg(-30)
        beta_angle = AngleDeg(30)
        self.assertAlmostEqual(alpha_angle.cos(), 0.8660254038)
        self.assertAlmostEqual(beta_angle.cos(), 0.8660254038)

    def test_sin(self):
        alpha_angle = AngleDeg(-30)
        beta_angle = AngleDeg(30)
        self.assertAlmostEqual(alpha_angle.sin(), -0.5)
        self.assertAlmostEqual(beta_angle.sin(), 0.5)


    def test_tan(self):
        alpha_angle = AngleDeg(-30)
        beta_angle = AngleDeg(30)
        self.assertAlmostEqual(alpha_angle.tan(), -0.5773502692)
        self.assertAlmostEqual(beta_angle.tan(),0.5773502692)

    def test_rad2deg(self):
        self.assertAlmostEqual(AngleDeg.rad2deg(0.52359877), 30 , 5)

    def test_deg2rad(self):
        self.assertAlmostEqual(AngleDeg.deg2rad(30), 0.523599 , 6)

    def test_cos_deg(self):
        self.assertAlmostEqual(AngleDeg.cos_deg(30), 0.8660254038)

    def test_sin_deg(self):
        self.assertAlmostEqual(AngleDeg.sin_deg(30), 0.5)

    def test_tan_deg(self):
        self.assertAlmostEqual(AngleDeg.tan_deg(30), 0.5773502692)

    def test_acos_deg(self):
        self.assertAlmostEqual(AngleDeg.acos_deg(0.8660254038), 30)

    def test_asin_deg(self):
        self.assertAlmostEqual(AngleDeg.asin_deg(0.5), 30)

    def test_atan_deg(self):
        self.assertAlmostEqual(AngleDeg.atan_deg(0.5773502692), 30)

    def test_atan2_deg(self):
        self.assertAlmostEqual(AngleDeg.atan2_deg(10,10), 45)

    def test_bisect(self):
        alpha_angle = AngleDeg(30)
        beta_angle = AngleDeg(0)
        gamma_angle = AngleDeg(15)
        self.assertEqual(AngleDeg.bisect(alpha_angle,beta_angle), gamma_angle)



""" test_matrix_2d.py file
    to test pyrusgeom Matrix2D class
"""
from unittest import TestCase
from pyrusgeom.matrix_2d import Matrix2D
from pyrusgeom.matrix_2d import Vector2D


class TestMatrix2D(TestCase):
    """TestMatrix2D class

    Args:
        TestCase (UnitTest): fail if any of tests falis
    """

    def test_matrix_init_assign_reset(self):
        mat_0 = Matrix2D()
        mat_1 = Matrix2D(1, 0, 0, 1, 0, 0)
        mat_2 = Matrix2D(1, 1, 1, 1, 1, 1)
        self.assertEqual(mat_0, mat_1)
        self.assertNotEqual(mat_0, mat_2)
        mat_2.reset()
        self.assertEqual(mat_0, mat_2)

    def test_det_invertible(self):
        mat_0 = Matrix2D()
        mat_1 = Matrix2D(2, 0, 2, 2, 2, 1)
        mat_2 = Matrix2D(5, 4, 5, 4, 0, 0)
        self.assertEqual(mat_0.det(), 1)
        self.assertEqual(mat_1.det(), 4)
        self.assertEqual(mat_2.det(), 0)
        self.assertTrue(mat_0.invertible())
        self.assertTrue(mat_1.invertible())
        self.assertFalse(mat_2.invertible())

    def test_inverted(self):
        mat_0 = Matrix2D()
        mat_1 = Matrix2D(2.0, 0.0, 2.0, 2.0, 2.0, 1.0)
        mat_2 = Matrix2D(5, 4, 5, 4, 0, 0)
        self.assertEqual(mat_0.inverted(), mat_0)
        self.assertTrue(mat_1.inverted(), Matrix2D(0.5, 0.0, -0.5, 0.5, -1.0, 0.5))
        self.assertEqual(mat_2.inverted(), Matrix2D(1.0, 0.0, 0.0, 1.0, 0.0, 0.0))

    def test_tranform(self):
        mat_0 = Matrix2D()
        mat_1 = Matrix2D(2.0, 0.0, 2.0, 2.0, 2.0, 1.0)
        vec_1 = Vector2D(10,10)
        self.assertEqual(mat_0.transform(vec_1),vec_1)
        self.assertEqual(mat_1.transform(vec_1),Vector2D(22,41))

    def test_mul(self):
        mat_0 = Matrix2D()
        mat_1 = mat_0 * mat_0
        # mat_1 = Matrix2D(2.0, 0.0, 2.0, 2.0, 2.0, 1.0)
        self.assertEqual(mat_1, Matrix2D())



"""soccer_math file
   math utility depending on RCSSServer2D
   TODO: use math_values
"""
from __future__ import annotations
# from typing import Union
import math

from pyrusgeom.vector_2d import Vector2D

EPS = 1.0e-8
SERVER_EPS = 1.0e-10


def kick_rate(dist: float, dir_diff: float, kick_rate_power: float,
              ball_size: float, player_size: float, kickable_margin: float) -> float:
    """calculate kick rate

    Args:
        dist (float): distance from player to ball
        dir_diff (float): angle difference from player's body to ball
        kick_rate_power (float): player's kick power rate parameter
        ball_size (float): ball radius
        player_size (float): player radius
        kickable_margin (float): player's kickable area margin

    Returns:
        float: rate of the kick power effect
    """
    return kick_rate_power * (1.0 - 0.25 * math.fabs(dir_diff)
                              / 180.0 - 0.25 * (dist - ball_size - player_size) / kickable_margin)


def dir_rate(rel_dash_dir: float, back_dash_rate: float, side_dash_rate: float) -> float:
    """calculate effective dash power rate according to the input dash direction

    Args:
        rel_dash_dir (float): relative dash direction
        back_dash_rate (float): server parameter
        side_dash_rate (float): server parameter

    Returns:
        float: dash power rate
    """
    if math.fabs(rel_dash_dir) > 90.0:
        return back_dash_rate - ((back_dash_rate - side_dash_rate) * (
            1.0 - (math.fabs(rel_dash_dir) - 90.0) / 90.0))
    return side_dash_rate + ((1.0 - side_dash_rate) * (1.0 - math.fabs(rel_dash_dir) / 90.0))


def effective_turn(turn_moment: float, speed: float, inertia_moment: float) -> float:
    """calculate effective turn moment.

    it may be useful to redefine self algorithm in movement action module

    Args:
        turn_moment (float): value used by turn command
        speed (float): player's current speed
        inertia_moment (float): player's inertia moment parameter

    Returns:
        float: calculated actual turn angle
    """
    return turn_moment / (1.0 + inertia_moment * speed)


def final_speed(dash_power: float, player_dash_power_rate: float,
                effort: float, decay: float) -> float:
    """calculate converged max speed, using "dash_power"

    NOTE: returned value should be compared with player_speed_max parameter

    Args:
        dash_power (float): value used by dash command
        player_dash_power_rate (float): player's dash power rate parameter
        effort (float): player's effort parameter
        decay (float): player's decay parameter

    Returns:
        float: final speed
    """
    return (math.fabs(dash_power) * player_dash_power_rate * effort) / (1.0 - decay)


def can_over_speed_max(dash_power: float, dash_power_rate: float, effort: float,
                       decay: float, speed_max: float) -> bool:
    """check if player's potential max speed is over player_speed_max parameter.

    Args:
        dash_power (float): value used by dash command
        dash_power_rate (float): player's dash power rate parameter
        effort (float): player's effort parameter
        decay (float): decay parameter
        speed_max (float): player's limited speed parameter

    Returns:
        bool: True if player can over player_speed_max. else False.
    """
    return math.fabs(dash_power) * dash_power_rate * effort > speed_max * (1.0 - decay)


def inertia_n_step_travel(initial_vel: Vector2D, n_step: int, decay: float) -> Vector2D:
    """estimate future travel after n steps only by inertia.

    No additional acceleration.

    Args:
        initial_vel (Vector2D): object's first velocity
        n_step (int): number of total steps
        decay (float): object's decay parameter

    Returns:
        Vector2D: vector of total travel
    """
    return (Vector2D(initial_vel.x(), initial_vel.y()) *
           ((1.0 - math.pow(decay, n_step)) / (1.0 - decay)))


def inertia_n_step_point(initial_pos: Vector2D,
                         initial_vel: Vector2D, n_step: int, decay: float) -> Vector2D:
    """estimate future point after n steps only by inertia.

    No additional acceleration.

    Args:
        initial_pos (Vector2D): object's first position
        initial_vel (Vector2D): object's first velocity
        n_step (int): number of total steps
        decay (float): object's decay parameter

    Returns:
        Vector2D: coordinate of the reached point
    """
    return (Vector2D(initial_pos.x(), initial_pos.y()) +
            inertia_n_step_travel(initial_vel, n_step, decay))


def inertia_n_step_distance(initial_speed: float, n_step: int, decay: float) -> float:
    """estimate travel distance only by inertia.

    No additional acceleration.

    Args:
        initial_speed (float): object's first speed
        n_step (int): number of total steps
        decay (float): decay object's decay parameter

    Returns:
        float: total travel distance
    """
    return initial_speed * (1.0 - math.pow(decay, n_step)) / (1.0 - decay)

def inertia_final_travel(initial_vel: Vector2D, decay: float) -> Vector2D:
    """calculate total travel only by inertia movement.

    Args:
        initial_vel (Vector2D): initial_vel object's first velocity
        decay (float): object's decay parameter

    Returns:
        Vector2D: final travel vector
    """
    return  Vector2D(initial_vel.x(), initial_vel.y()) / (1.0 - decay)


def inertia_final_point(initial_pos: Vector2D, initial_vel: Vector2D, decay: float) -> Vector2D:
    """calculate final reach point only by inertia.

    Args:
        initial_pos (Vector2D): object's first position
        initial_vel (Vector2D): object's first velocity
        decay (float): decay parameter

    Returns:
        Vector2D: coordinate of the reached point
    """
    return Vector2D(initial_pos.x(), initial_pos.y()) + inertia_final_travel(initial_vel, decay)


def inertia_final_distance(initial_speed: float, decay: float) -> float:
    """calculate total travel distance only by inertia.

    Args:
        initial_speed (float): object's first speed
        decay (float): object's decay parameter

    Returns:
        float: distance value that the object reaches
    """
    return initial_speed / (1.0 - decay)


def r_int(floating_point: float) -> int:
    """Rounds the floating-point argument f to an integer value (in floating-point format),
    using the current rounding mode.

    Args:
        floating_point (float): floating point value

    Returns:
        int: If no errors occur, the nearest integer value to f, according to the current
    rounding mode, is returned.
    """

    fi_og = int(floating_point)
    fi_diff = math.fabs(fi_og - floating_point)
    fi_left = fi_og - 1
    fi_left_diff = math.fabs(fi_left - floating_point)
    fi_right = fi_og + 1
    fi_right_diff = math.fabs(fi_right - floating_point)
    if fi_diff < fi_left_diff and fi_diff < fi_right_diff:
        return fi_og
    if fi_left_diff < fi_right_diff:
        return fi_left
    return fi_right


def calc_length_geom_series(first_term: float, sum_all: float, ratio: float) -> float:
    """caluculate the length of a geometric series

    Args:
        first_term (float): value of the first term
        sum_all (float): sum of a geometric series
        ratio (float): multiplication ratio

    Returns:
        float: a round number of the length of geometric series
    """
    if first_term <= SERVER_EPS or sum_all < 0.0 or ratio <= SERVER_EPS:
        return -1.0
    if sum_all <= SERVER_EPS:
        return 0.0
    tmp = 1.0 + sum_all * (ratio - 1.0) / first_term
    if tmp <= SERVER_EPS:
        return -1.0
    return math.log(tmp) / math.log(ratio)

def calc_first_term_geom_series(sums: float, ratio: float, length: int) -> float:
    """calculate the frist term of geometric series

    Args:
        sums (float): sum of a geometric series
        ratio (float): ratio of series
        length (int): length of a geometric series

    Returns:
        float: frist term of geometric series
    """
    return sums * (1.0 - ratio) / (1.0 - math.pow(ratio, length))


def bound(num_0: float, num_1: float, num_2: float) -> float:
    """calc the bound between three input float

    Args:
        num_0 (float): 1st number
        num_1 (float): 2nd number
        num_2 (float): 3rd number

    Returns:
        float: bounded number
    """
    return min(max(num_0, num_1), num_2)

def min_max(low: float, number: float, high: float) -> float:
    """min max the input number

    Args:
        low (float): lower bound
        number (float): input number
        high (float): higher bound

    Returns:
        float: bounded number
    """
    return min(max(low, number), high)


def f_range(start: float, stop: float = None, step: float = None) -> float:
    """Use float number in range() function

    if stop and step argument is null set start=0.0 and step = 1.0

    Args:
        start (float): the start number
        stop (float, optional): the end number. Defaults to None.
        step (float, optional): start += step until end. Defaults to None.


    Args:
        start (float): [description]
        stop (float, optional): [description]. Defaults to None.
        step (float, optional): [description]. Defaults to None.

    Returns:
        float: number

    Yields:
        Iterator[float]: [description]
    """
    if stop is None:
        stop = start + 0.0
        start = 0.0
    if step is None:
        step = 1.0
    while True:
        if step > 0 and start >= stop:
            break
        if step < 0 and start <= stop:
            break
        yield start
        start = start + step

# localization

def quantize(value: float, qstep: float) -> float:
    """quantize a floating point number
    same as define Quantize(v,q) ((rint((v)/(q)))*(q))

    Args:
        value (float):  value to be rounded
        qstep (float): round precision

    Returns:
        float: rounded value
    """
    return r_int(value / qstep) * qstep


def quantize_dist(unq_dist: float, qstep: float) -> float:
    """calculate quantized distance value about dist quantization

    Args:
        unq_dist (float): actual distance
        qstep (float): server parameter

    Returns:
        float: quantized distance
    """
    return quantize(math.exp(quantize(math.log(unq_dist + SERVER_EPS), qstep)), 0.1)


def unquantize_min(dist: float, qstep: float) -> float:
    """calculate minimal value by inverse quantize function

    Args:
        dist (float): quantized distance
        qstep (float): server parameter

    Returns:
        float: minimal distance within un quantized distance range
    """
    return (r_int(dist / qstep) - 0.5) * qstep


def unquantize_max(dist: float, qstep: float) -> float:
    """calculate maximal value by inverse quantize function

    Args:
        dist (float): dist quantized distance
        qstep (float): server parameter

    Returns:
        float: maximal distance within un quantized distance range
    """
    return (r_int(dist / qstep) + 0.5) * qstep


def wind_effect(speed: float, weight: float,
                wind_force: float, wind_dir: float,
                wind_weight: float, wind_rand: float,
                wind_error: Vector2D) -> Vector2D:
    """calculate wind effect

    Args:
        speed (float): current object's speed
        weight (float): object's speed weight
        wind_force (float): server parameter
        wind_dir (float): server parameter
        wind_weight (float): server parameter
        wind_rand (float): server parameter
        wind_error (Vector2D): error value that is calculated by this method

    Returns:
        Vector2D: wind effect acceleration
    """
    wind_vec = Vector2D()
    wind_vec.polar2vector(wind_force, wind_dir)

    if wind_error:
        wind_error.assign(speed * wind_vec.x() * wind_rand
                          / (weight * wind_weight),
                          speed * wind_vec.y() * wind_rand
                          / (weight * wind_weight))

    return Vector2D(speed * wind_vec.x() / (weight * wind_weight),
                    speed * wind_vec.y() / (weight * wind_weight))


def unquantize_error(see_dist: float, qstep: float) -> float:
    """calculate min max error range by inverse quantize function

    Args:
        see_dist (float): seen(quantized) distance
        qstep (float): server parameter

    Returns:
        float: error value of inverse un quantized distance
    """
    min_dist = (math.exp(unquantize_min(
        math.log(unquantize_min(see_dist, 0.1)), qstep)) - SERVER_EPS)
    max_dist = (math.exp(unquantize_max(
        math.log(unquantize_max(see_dist, 0.1)), qstep)) - SERVER_EPS)
    return math.fabs(max_dist - min_dist)
